"""The main system dynamics model, handles tracking and running
stock and flow equations to run simulation(s)."""

import json
import math
import threading
import warnings
from copy import deepcopy
from typing import Any

import arviz as az
import numpy as np
import pymc as pm
import xarray as xr
from graphviz import Digraph, set_jupyter_format
from pytensor import compile
from tqdm.auto import tqdm

import reno


class _ModelContexts(threading.local):
    """Similar to how PyMC manages this, we keep a model manager to allow models
    to be used as context managers, so a model name of ``my_really_long_model``
    doesn't have to be repeated over and over and over when defining all of the
    stocks and flows on it."""

    def __init__(self):
        self.current_models: list[Model] = []

    @property
    def current_model(self):
        if len(self.current_models) > 0:
            return self.current_models[-1]
        return None

    # TODO: is a parent_model property necessary?


# similar to PyMC, single global (thread-safe) list of current model contexts
# (e.g. within a with block)
MODEL_CONTEXTS = _ModelContexts()


class Model:
    """Class for a distinct, simulatable set of related stocks and flows.

    The expectation is to create a model instance, and then assign stocks,
    flows, variables, and submodels as attributes on that instance.

    Args:
        name (str): Optional name to give the model, should be used when
            submodels are in play, as the model name is used to help
            visually distinguish which model things belong to.
        n (int): The number of samples to simulate at once (by default).
        steps (int): How many time steps to run the simulation for (by default).
        label (str): Optional visual label to use when printing model things if
            cleaner than using the name.
        doc (str): Optional docstring to explain/describe the model.

    Example:
        >>> import reno

        >>> tub = reno.Model("Tub model")
        >>> tub.water_level = reno.Stock()
        >>> tub.faucet = reno.Flow(5)
        >>> tub.drain = reno.Flow(tub.water_level / 2)

        >>> tub.water_level += tub.faucet
        >>> tub.water_level -= tub.drain

        >>> tub(steps=50)
    """

    def __init__(
        self,
        name: str = None,
        n: int = 1,
        steps: int = 10,
        label: str = None,
        doc: str = None,
    ):
        self.stocks: list[reno.Stock] = []
        """List of all stocks associated with this model - don't modify directly,
        assign stocks as attributes directly on model (e.g. ``model.my_stock = reno.Stock()``)
        """
        self.flows: list[reno.Flow] = []
        """List of all flows associated with this model - don't modify directly,
        assign flows as attributes directly on model (e.g. ``model.my_flow = reno.Flow()``)
        """
        self.vars: list[reno.Variable] = []
        """List of all variables associated with this model - don't modify directly,
        assign vars as attributes directly on model (e.g. ``model.my_var = reno.Variable()``)
        """
        self.metrics: list[reno.components.Metric] = []
        """List of all metrics associated with this model - don't modify directly,
        assign metrics as attributes directly on model (e.g. ``model.my_metric = reno.Metric()``)
        """
        self.models: list[reno.Model] = []
        """List of submodels of this model - don't modify directly, assign submodels as
        attributes directly on model (e.g. ``model.my_submodel = reno.Model()``
        """

        self.parent: reno.Model = None
        """Parent model if applicable. This gets set in the __setattr__ when a submodel
        is assigned."""

        self.name: str = name
        if label is None:
            label = name
        self.label: str = label
        """String label to use in any visualizations/outputs to refer to this model. defaults
        to name if not used."""

        # not sure if I need these separately or not? Just thinking it
        # might make it easier to be able to set than to have to always pass in
        # simulate
        self.steps: int = steps
        """The number of timesteps to use by default in this model's simulations, can
        be overriden per model call."""
        self.n: int = n
        """The number of simulations to run in parallel by default, can be overriden
        per model call."""

        # previous n/steps variables to revert after a particular run.
        self.last_n: int = n
        self.last_steps: int = steps

        self.doc: str = doc
        """Docstring/comment to describe this model."""
        # TODO: offset?

        self.trace = None  # assigned by pymc()
        """Arviz trace produced by the last pymc run. This is used for plotting
        defaults on functions that aren't explicitly passed traces."""
        self.trace_RVs: list[str] = []
        """The set of string names of any random variables from the last pymc run."""

        self._unnamed_references: list = []
        """TrackedReferences added within a context manager go here so names can be
        assigned based on local frame variable names when it exits."""

        self.group_colors: dict[str, str] = {}
        """Specify colors for groups/cgroups, group name is the key."""

        self.default_hide_groups: list[str] = []
        """If there are any groups/cgroups that shouldn't be displayed in the stock
        flow diagram by default, list them here."""

        if Model.get_context() is not None:
            Model.get_context()._unnamed_references.append(self)

    @property
    def groups(self) -> dict[str, "reno.components.TrackedReference"]:
        """Get the list of groups and cgroups in the model. These can
        be used to control colors of components in stock/flow diagrams
        (see ``model.group_color``) and default show/hide behavior (see
        ``model.default_hide_groups``)"""
        group_list = {}
        for ref in self.all_refs():
            if ref.group != "":
                if ref.group not in group_list:
                    group_list[ref.group] = []
                group_list[ref.group].append(ref)
            if ref.cgroup != "":
                if isinstance(ref.cgroup, list):
                    for cgroup in ref.cgroup:
                        if ref.cgroup not in group_list:
                            group_list[ref.cgroup] = []
                        group_list[ref.cgroup].append(ref)
                else:
                    if ref.cgroup not in group_list:
                        group_list[ref.cgroup] = []
                    group_list[ref.cgroup].append(ref)
        return group_list

    @classmethod
    def get_context(cls):
        return MODEL_CONTEXTS.current_model

    def __enter__(self):
        MODEL_CONTEXTS.current_models.append(self)

    def __exit__(self, exc_type, exc_value, traceback):
        MODEL_CONTEXTS.current_models.pop()
        for index, ref in enumerate(self._unnamed_references):
            # make sure adding this ref wasn't already handled
            if isinstance(
                ref, (reno.components.TrackedReference, reno.components.Metric)
            ):
                if ref.name is not None and ref.model is not None:
                    continue
            elif isinstance(ref, reno.components.TrackedReference):
                if ref.name is not None and ref.parent is not None:
                    continue
            name = reno.utils._get_assigned_var_name(ref)
            if name is None and ref.implicit:
                if hasattr(ref, "_implicit_target"):
                    name = f"_{ref._implicit_target.qual_name()}_inflow_{ref._implicit_target_index}"
                else:
                    name = "_implicit_ref_" + str(id(ref))
            setattr(self, name, self._unnamed_references[index])
        self._unnamed_references = []

    def _is_init_ref(self, name: str) -> bool:
        """Is the name of the stock/flow/variable reference an initial value reference?
        This allows referring to a reference's init with ``my_model.my_reference_0``.
        """
        if name.endswith("_0") and hasattr(self, name.removesuffix("_0")):
            return True
        return False

    def __setattr__(self, name, value):
        # It's easy to accidentally overwrite a reference with its equation (if
        # you forget to do `ref.eq`). Check for this behavior by checking to see
        # if the name being set _already exists_ and is a Flow/Variable, and
        # if so throw a warning
        if hasattr(self, name):
            if isinstance(
                getattr(self, name),
                (reno.components.Flow, reno.components.Variable),
            ):
                warnings.warn(
                    f"Reassigning entire model reference `{name}` with an equation, did you mean to instead assign the reference's equation, e.g. `{name}.eq = ...`?",
                    RuntimeWarning,
                )

        # Any refs/metrics/models assigned as an attribute should also be added
        # to the appropriate tracking lists
        if isinstance(value, reno.components.Stock) and value not in self.stocks:
            value.model = self
            self.stocks.append(value)
        elif isinstance(value, reno.components.Flow) and value not in self.flows:
            value.model = self
            self.flows.append(value)
        elif isinstance(value, reno.components.Variable) and value not in self.vars:
            value.model = self
            self.vars.append(value)
        elif isinstance(value, reno.components.Metric) and value not in self.metrics:
            value.model = self
            self.metrics.append(value)
        elif isinstance(value, Model) and not name == "parent" and not name == "model":
            # TODO: do we need to make a copy?
            value.parent = self
            self.models.append(value)

        # allow more ""organic"" assigning of init values with `my_model.my_stock_0 = 5`
        if self._is_init_ref(name):
            getattr(self, name.removesuffix("_0")).init = value
            return

        # Use the reference's name as its label in visual displays unless
        # otherwise requested. (e.g. if label is sepcified)
        # also set the reference's name to be the attribute you're setting as on
        # this model.
        if isinstance(
            value,
            (
                reno.components.Stock,
                reno.components.Flow,
                reno.components.Variable,
                reno.components.Metric,
                Model,
            ),
        ) and not (isinstance(value, Model) and (name == "parent" or name == "model")):
            # if isinstance(value, reno.components.Flow):
            #     if value.name is None and value.implicit:
            #         name = "_implicit_ref_" + str(id(value))
            if value.label is None or value.label == value.name:
                value.label = name
            value.name = name

        object.__setattr__(self, name, value)

    def __getattr__(self, name):
        # allow more ""organic"" retrieval of init values, e.g. `my_model.my_stock_0`
        if name.endswith("_0"):
            if hasattr(self, name.removesuffix("_0")):
                return getattr(self, name.removesuffix("_0")).init
            raise AttributeError(name.removesuffix("_0") + " not found")
        if "." in name:
            # check for submodel (this works recursively)
            parts = name.split(".")
            if hasattr(self, parts[0]):
                return getattr(getattr(self, parts[0]), ".".join(parts[1:]))
        else:
            # NOTE: without this here, deepcopy fails, see
            # https://python-list.python.narkive.com/pxEGTJtL/folks-what-s-wrong-with-this
            raise AttributeError(name + " not found")

    def add(self, name: str, value: "reno.components.Reference | reno.Model"):
        """Add the passed tracked reference to the model with the provided name. This is used for
        programmatically adding stocks/flows in a context where the name is dynamically created and
        you can't simply ``model.my_name = ``."""
        # this is effectively just an alias for __setattr__ but I don't want
        # people to have to manually call model.__setattr__
        setattr(self, name, value)

    def all_stocks(self) -> list:
        """Get all stocks from this and all submodels."""
        full_list = [*self.stocks]
        for submodel in self.models:
            full_list.extend(submodel.all_stocks())
        return full_list

    def all_flows(self) -> list:
        """Get all flows from this and all submodels."""
        full_list = [*self.flows]
        for submodel in self.models:
            full_list.extend(submodel.all_flows())
        return full_list

    def all_vars(self) -> list:
        """Get all vars from this and all submodels."""
        full_list = [*self.vars]
        for submodel in self.models:
            full_list.extend(submodel.all_vars())
        return full_list

    def all_refs(self) -> list:
        """Get all stocks, flows, and vars, from this and all submodels in one giant
        ordered list (does not include metrics, use ``all_metrics()`` separately.)"""
        return self.all_stocks() + self.all_flows() + self.all_vars()

    def all_metrics(self) -> list:
        """Get all metrics recursively"""
        full_list = [*self.metrics]
        for submodel in self.models:
            full_list.extend(submodel.all_metrics())
        return full_list

    def _reset_type_and_shape_info(self):
        for ref in self.all_refs():
            ref._shape = None
            ref._dtype = None

    def _recursive_sub_populate_n_steps(self, n: int, steps: int):
        """Recursively populate all submodel's n/steps settings."""
        # TODO: this feels like it shouldn't be necessary and also doesn't
        # handle if a ref is in another model that wasn't explicitly set as a
        # submodel of this one?
        self.last_n = n
        self.last_steps = steps
        for model in self.models:
            model._recursive_sub_populate_n_steps(n, steps)

    def _populate(self, n: int, steps: int):
        """Initialize all tracked references with appropriately sized numpy
        matrices."""
        self._reset_type_and_shape_info()
        self._find_all_extended_op_implicit_components()
        self._recursive_sub_populate_n_steps(n, steps)

        ref_compute_order = self.dependency_compute_order(inits_order=True)

        for ref in ref_compute_order:
            ref.populate(n, steps)

    def _find_all_extended_op_implicit_components(self):
        """Go through every equation and assign any implicit components from
        extended operations."""
        extended_ops = []
        assoc_refs = []  # parallel list with which ref the correspo op came from
        # (not necessarily accurate, but we need an easy way to be able to name
        # separately)
        assoc_ref_counter = {}  # parallel list with how many times this ref is used
        # (useful for indexing the sub components to ensure no name conflicts)
        for ref in self.flows + self.vars:
            found_ops = ref.find_refs_of_type(reno.components.ExtendedOperation)
            for op in found_ops:
                if op not in extended_ops:
                    extended_ops.append(op)
                    assoc_refs.append(ref)
                    assoc_ref_counter[ref] = 0

        for i, op in enumerate(extended_ops):
            # NOTE: keep this above, we don't want the name to be _0 or we get
            # bad things because that would otherwise reference an init
            assoc_ref_counter[assoc_refs[i]] += 1
            for key, component in op.implicit_components.items():
                name = f"{assoc_refs[i].qual_name()}_{key}_{assoc_ref_counter[assoc_refs[i]]}"
                if hasattr(self, name):
                    delattr(self, name)
                self.add(name, component)

        # recursively do this for submodels too
        for model in self.models:
            model._find_all_extended_op_implicit_components()

    def simulator(
        self, n: int = None, steps: int = None, quiet: bool = False, debug: bool = False
    ):
        """An iterator to use for running the simulation step by step. Leaving n and/or
        steps None will use the model's default (as defined in constructor.)"""
        if n is None:
            n = self.n
        if steps is None:
            steps = self.steps

        self._recursive_sub_populate_n_steps(
            n, steps
        )  # TODO: (2025.07.28) isn't this redundant?
        # (should already be being handled in _populate?)
        self._populate(n, steps)

        # note that dependency_compute_order includes all submodels' refs
        ref_compute_order = self.dependency_compute_order(inits_order=False)

        for step in tqdm(range(1, steps), disable=quiet, desc=self.name):
            if debug:
                print("Beginning step", step, self.name)
            for ref in ref_compute_order:
                ref.eval(step, save=True)
            yield

        self.run_metrics(n, steps)

    def simulate(
        self, n: int = None, steps: int = None, quiet: bool = False, debug: bool = False
    ):
        """Run each step of the the full simulation. Leaving n and/or
        steps None will use the model's default (as defined in constructor.)"""
        for step in self.simulator(n, steps, quiet, debug):
            pass

    def run_metrics(self, n: int = None, steps: int = None):
        """Run all metric equations on a completed simulation. Calling this
        function assumes the full simulation has already run."""
        if n is None:
            n = self.n
        if steps is None:
            steps = self.steps

        metrics = reno.utils.dependency_compute_order([*self.all_metrics()])

        # populate any metrics (only flags apply to every timestep currently)
        for metric in metrics:
            if isinstance(metric, reno.components.Flag):
                metric.populate(n, steps)

        # compute flag boolean values for each sample at each timestep
        for step in range(1, steps):
            for metric in metrics:
                if isinstance(metric, reno.components.Flag):
                    metric.eval(step, True)

        # run any postmeasurement equations, usually 1 per sample
        for metric in metrics:
            if isinstance(metric, reno.components.Metric):
                metric.eval(steps - 1, True)
                # TODO: is - 1 correct? I don't think it is

    def graph(
        self,
        show_vars: bool = True,
        exclude_vars: list[str] = None,
        sparklines: bool = False,
        sparkdensities: bool = False,
        sparkall: bool = False,
        g: Digraph = None,
        traces: list[xr.Dataset] = None,
        universe: list[reno.components.TrackedReference] = None,
        lr: bool = False,
        hide_groups: list[str] = None,
        show_groups: list[str] = None,
        group_colors: dict[str | tuple["reno.components.TrackedReference"], str] = None,
    ) -> Digraph:
        """Generate a graphviz dot graph for all the stocks and flows of the passed model,
        optionally including sparklines if a simulation has been run.

        Args:
            model (reno.model.Model): The model to collect stocks/flows/variables from.
            show_vars (bool): Whether to render variables in the diagram, or just stocks and flows
                (for very complex models, hiding variables can make it a bit easier to visually
                parse.)
            exclude_var_names (list[str]): Specific variables to hide in the diagram, can be used
                with ``show_vars=True`` to just show specific variables of interest.
            sparklines (bool): Draw mini graphs to the right of each stock showing plotting their
                values through time. This assumes the model has either been run, or traces are passed
                in manually with the ``traces`` argument.
            sparkdensities (bool): Draw mini density plots/histograms next to any variables that
                sample from distributions. This assumes the model has either been run, or traces
                are passed in manually with the ``traces`` argument.
            g (Digraph): Graphviz Digraph instance to render the nodes/edges on. Mostly only for
                internal use for drawing subgraphs for submodels.
            traces (list[xr.Dataset]): A list of traces or model run datasets to use for drawing
                spark plots. Each dataset will be rendered in a different color.
            universe (list[TrackedReference]): Limit rendered nodes to only those listed here, this
                includes all of stocks/flows/variables. (This acts as an initial filter, ``show_vars``
                and ``exclude_var_names`` still applies after this.)
            lr (bool): By default the graphviz plot tries to orient top-down. Specify ``True`` to
                try to orient it left-right.
            hide_groups (list[str]): A list of group/cgroup names to hide during diagramming, overriding
                model.default_hide_groups.
            show_groups (list[str]): A list of group/cgroup names to show during diagramming, overriding
                model.default_hide_groups.
            group_colors dict[str | tuple["reno.components.TrackedReference"], str]: Dictionary specifying
                colors to render groups with. An ad-hoc group defined by a tuple of references can also be
                used as a key if the appropriate cgroup does not already exist on the references.

        Returns:
            The populated Digraph instance (Jupyter can natively render this in a cell output.)
        """
        if exclude_vars is None:
            exclude_vars = []
        set_jupyter_format("png")
        diagram, _ = reno.diagrams.stock_flow_diagram(
            self,
            show_vars,
            exclude_vars,
            sparklines,
            sparkdensities,
            sparkall,
            g,
            traces,
            universe,
            lr,
            hide_groups,
            show_groups,
            group_colors,
        )
        return diagram

    def latex(
        self, docs: bool = True, t: int = None, sample: int = 0, raw_str: bool = False
    ):
        """Get an interactive latex ipywidget listing all of the equations in system. Each equation
        line is clickable, clicking will highlight where else in the system that equation's result
        is being used.

        Args:
            docs (bool): Whether to include the documentation string beneath each equation or not.
            t (int): If specified, show the values of every reference at the specified timestep (note
                that stock values will be from the previous timestep)
            sample (int): Which sample (row) to show the values from if ``t`` was specified.
            raw_str (bool): Set this to True to just get the string of latex instead of
                the interactive widget.
        """
        debug = False
        if t is not None:
            debug = True

        latex_obj = reno.viz.ModelLatex(self, docs, t=t, sample=sample, debug=debug)
        if raw_str:
            return latex_obj.latex.data
        return latex_obj.widget

    def plot_stocks(self, cols: int = None, rows: int = None, **figargs):
        """Shortcut function to quickly get a set of graphs for each stock."""
        return reno.viz.plot_refs(self.stocks, cols=cols, rows=rows, **figargs)

    def copy(self, name: str = None) -> "Model":
        """Make a separate copy of this model with the desired name.

        This is useful for making a separate model instance that you can modify the
        equations of for comparison against the original.
        """
        new_model = deepcopy(self)
        if name is not None:
            new_model.name = name
        return new_model

    def free_refs(self, recursive: bool = False):
        """Get all free "variables" (not component variables) for this model, or things that
        aren't defined in terms of anything else.

        A free variable is a variable with no references to other variables, stocks or flows.
        """
        free = []
        inits = []

        # TODO: the ref in seek_refs check probably then needs to see if the
        # init function is a scalar/dist/int/float, and similarly expand
        # conditional eq checks to include int/float

        if not recursive:
            # if not directly recursing, list each submodel itself as the parameter
            for model in self.models:
                free.append(model)
        else:
            # otherwise find and list each of the free variables within each submodel
            for model in self.models:
                # print(model.name)
                # for ref in model.free_refs(recursive):
                #     print(ref, type(ref))
                free.extend(
                    [
                        f"{model.name}.{ref}"
                        for ref in model.free_refs(recursive)
                        # if not ref.implicit
                    ]
                )

        # a variable is free if its equation is directly a scalar or distribution, or hasn't
        # been assigned an equation. Flows use same logic as variables here.
        for ref in self.vars + self.flows:
            if ref.implicit:
                continue

            if reno.utils.is_free_var(ref.eq):
                free.append(ref)

            # if a variable/flow is a function defined in terms of itself, needs an init.
            if ref in ref.seek_refs():
                inits.append(ref)

        # stock inits (_0) are normally free, would be weird if init was defined
        # in terms of another var?
        for stock in self.stocks:
            if reno.utils.is_free_var(stock.init):
                inits.append(stock)

        return [ref.name if hasattr(ref, "name") else ref for ref in free] + [
            f"{ref.name}_0" for ref in inits
        ]

    def get_docs(self, as_dict: bool = False) -> str:
        """Get a full docstring for using this model as a function, maybe useful for
        allowing a model to be used as a tool for LLMs?

        Use as_dict for submodels, so example string shows args as dictionary."""
        free_refs = self.free_refs()

        docstring = self.doc if self.doc is not None else ""

        default_args = ""
        arg_strings = []
        for key, value in self.config().items():
            value_to_print = value
            if isinstance(value, reno.Scalar):
                value_to_print = value.value
            arg_strings.append(f"{key}={value_to_print}")
        default_args = ", ".join(arg_strings)

        docstring += f"\n\nExample:\n\t{self.name}({default_args})"

        docstring += "\n\nArgs:"

        for ref_name in free_refs:
            if self._is_init_ref(ref_name):
                ref = getattr(self, ref_name.removesuffix("_0"))
            else:
                ref = getattr(self, ref_name)
            doc = ref.doc

            docstring += f"\n\t{ref_name}"
            if doc is not None:
                if self._is_init_ref(ref_name):
                    docstring += f": Initial value for {doc[:1].lower()}{doc[1:]}"
                else:
                    docstring += f": {doc}"

        return docstring

    def config(self, **free_refs) -> dict:
        """Get/set model configuration. This function allows specifying one or more
        free variables - anything not set uses the default.

        Returns:
            The resulting configuration dictionary, with the free ref names as keys.
        """
        available_free_refs = self.free_refs()

        # track all free variables for this sim, including those not explicitly
        # passed
        config = {}

        # initialize any vars/flows/stocks based on optionally passed references
        for ref_name in available_free_refs:
            # handle sub model configuration if applicable
            if isinstance(getattr(self, ref_name), Model):
                sub_model = getattr(self, ref_name)
                passed_sub_config = {}
                if ref_name in free_refs:
                    if not isinstance(free_refs[ref_name], dict):
                        raise AttributeError(
                            "sub model configuration must be passed as a sub dictionary"
                        )
                    passed_sub_config = free_refs[ref_name]
                sub_config = sub_model.config(**passed_sub_config)
                config[ref_name] = sub_config
            else:
                # any refs that _aren't_ an entire submodel
                if ref_name in free_refs:
                    # print(ref_name, "=", free_refs[ref_name])  # debug
                    if self._is_init_ref(ref_name):
                        init_eq_part = free_refs[ref_name]
                        if isinstance(init_eq_part, (int, float)):
                            init_eq_part = reno.components.Scalar(init_eq_part)
                        setattr(self, ref_name, init_eq_part)
                    else:
                        eq_part = free_refs[ref_name]
                        # TODO: dear lord there's almost certainly a cleaner way
                        # to do this...
                        if isinstance(eq_part, (int, float)):
                            eq_part = reno.components.Scalar(eq_part)
                        setattr(getattr(self, ref_name), "eq", eq_part)
                        # setattr(getattr(self, ref_name), "eq", free_refs[ref_name])
                    config[ref_name] = free_refs[ref_name]
                else:
                    val = getattr(self, ref_name)
                    # if val is None:
                    #     # val = 0  # saving this for posterity. BAD NATHAN. This
                    #     # causes asymmetric get vs set, e.g. I can't call
                    #     # config(**config()) without the end configuration being
                    #     # technically different. This was fine for Reno's
                    #     # underlying math system, but for PyMC where types
                    #     # become very important, it can accidentally convert
                    #     # what was initially assumed 0.0 to 0, (specifically
                    #     # this was causing stock _init values to become
                    #     # integers, completely screwing up the subsequent math)
                    if isinstance(val, reno.components.TrackedReference):
                        val = val.eq
                    config[ref_name] = val

        return config

    def get_nonrecursive_config(self):
        """Only get the free refs config from _this_ model, no submodels.

        Useful for assigning dataset attrs."""
        free_refs = self.free_refs()

        config = {}

        for ref_name in free_refs:
            if not isinstance(getattr(self, ref_name), Model):
                val = getattr(self, ref_name)
                if val is None:
                    val = 0
                if isinstance(val, reno.components.TrackedReference):
                    val = val.eq
                config[ref_name] = val

        return config

    def load_dataset(self, ds: xr.Dataset):
        """Take all the tracked ref sequence data and load them into the matching
        model's tracked refs stored values. This is useful for using ``.latex(t=...)``
        in debug mode for diving into a specific run."""
        unfound = []
        n = -1
        steps = -1

        # find n/steps
        for var in ds.variables:
            if len(ds.variables[var].shape) > 1:
                n = ds.variables[var].values.shape[0]
                steps = ds.variables[var].values.shape[1]
                break

        for ref in self.dependency_compute_order(inits_order=True) + self.metrics:
            found = False
            # find the corresponding variable in the dataset
            for var in list(ds.variables.keys()):
                # handle either . or _ parent model syntax in name
                # (pymc conversion uses _, default reno uses .)
                # NOTE: (2025-06-27) - switching to reno using _ as well, it's
                # too confusing otherwise.
                if var == ref.qual_name(False) or var == ref.qual_name(True):
                    if not isinstance(ref, reno.components.Metric):
                        ref.populate(
                            n, steps
                        )  # order shouldn't matter since replacing value anyway
                    ref.value = ds.variables[var].values
                    found = True
                    break

            if not found:
                unfound.append(ref)
                warnings.warn(
                    f"Loading values from dataset found no value for model reference '{ref.qual_name()}', value will be reset",
                    RuntimeWarning,
                )
                # TODO: order against dependency order?
                ref.populate(n, steps)

        # TODO: collect config as well?

    def dataset(self) -> xr.Dataset:
        """Turn all of the model's tracked reference values into an xarray dataset,
        including the "configuration" of the input parameters etc."""
        sub_dses = {}
        if len(self.models) > 0:
            for model in self.models:
                sub_dses[model.name] = model.dataset()

        # add non-static non-dim base refs (stocks, flows, vars)
        # partial case 1, (sample, t)  (see TrackedReference for cases
        # explanation)
        all_refs = self.stocks + self.flows + self.vars
        all_refs = [ref for ref in all_refs if not ref.implicit]
        ds = xr.Dataset(
            {
                # ref.name: (["sample", "step"], ref.value)
                ref.qual_name(): (["sample", "step"], ref.value)
                for ref in all_refs
                if not ref.is_static() and ref.dim == 1
            },
            coords={
                "sample": (["sample"], list(range(self.last_n))),
                "step": (["step"], list(range(self.last_steps))),
                # "vec": (["vec"], []),
            },
            attrs=self.get_nonrecursive_config(),
        )

        # handle any multi-dim refs
        multidim_vars = {}
        for ref in all_refs:
            if ref.dim > 1:
                vec_name = f"{ref.qual_name()}_vec"
                if ref.is_static():
                    if ref._sample_dim:
                        # case 2, (sample, dim)
                        dims = ["sample", vec_name]
                        coords = {
                            "sample": (["sample"], list(range(self.last_n))),
                            vec_name: ([vec_name], list(range(ref.dim))),
                        }
                        val = ref.value
                        if ref.value.shape[0] != self.last_n:
                            val = np.broadcast_to(ref.value, (self.last_n, ref.dim))
                    else:
                        # case 1, (dim,)
                        dims = ["sample", vec_name]
                        coords = {
                            "sample": (["sample"], list(range(self.last_n))),
                            vec_name: ([vec_name], list(range(ref.dim))),
                        }
                        val = ref.value
                        # TODO: I'm pretty sure this needs a repeat per last_n
                        # with axis=0...a broadcast may not work
                        val = np.broadcast_to(val, (self.last_n, ref.dim))
                else:
                    # partial case 1, (sample, t, dim)
                    dims = ["sample", "step", vec_name]
                    coords = {
                        "sample": (["sample"], list(range(self.last_n))),
                        "step": (["step"], list(range(self.last_steps))),
                        vec_name: ([vec_name], list(range(ref.dim))),
                    }
                    val = ref.value

                da = xr.DataArray(data=val, dims=dims, coords=coords)
                multidim_vars[ref.qual_name()] = da
        ds = ds.assign(multidim_vars)

        # handle any static refs (don't change wrt to step)
        # NOTE: this is only dealing with vectors right now, not single nums?
        static_refs = {}
        for ref in all_refs:
            if ref.is_static() and ref.dim == 1:
                val = ref.value
                # print(
                #     "(dataset)",
                #     ref.qual_name(),
                #     type(ref.value),
                #     ref.value.shape,
                #     ref.value,
                #     ref._static,
                #     ref._sample_dim,
                # )
                # TODO: geeeez there's got to be a better way?? np.min/max etc.
                # will return a numpy.int even if the inputs are not numpy
                # types (e.g. python int)
                if isinstance(ref.value, (int, float)) or (
                    isinstance(ref.value, np.ndarray) and len(ref.value.shape) == 0
                ):
                    val = np.broadcast_to(ref.value, (self.last_n,))
                elif len(ref.value.shape) > 0 and ref.value.shape[0] != self.last_n:
                    val = np.broadcast_to(ref.value, (self.last_n,))
                elif len(ref.value.shape) == 0:
                    val = np.broadcast_to(ref.value, (self.last_n,))
                static_refs[ref.qual_name()] = (["sample"], val)
        ds = ds.assign(static_refs)

        # add metrics, note that some metrics will be 1 per sample, others will
        # be 1 per step
        new_vars = {}
        for metric in self.metrics:
            if len(metric.value.shape) == 1:
                # 1 per sample
                coords = ["sample"]
            else:
                # 1 per step (e.g. flags)
                coords = ["sample", "step"]
            new_vars[metric.qual_name()] = (coords, metric.value)
        ds = ds.assign(new_vars)

        # merge in any sub datasets
        all_attrs = {}
        for sub_ds_name, sub_ds in sub_dses.items():
            # essentially flatten attr names
            renamed_attrs = {}
            for attr in sub_ds.attrs:
                renamed_attrs[f"{sub_ds_name}.{attr}"] = sub_ds.attrs[attr]
            sub_ds = sub_ds.drop_attrs(deep=False)
            sub_ds = sub_ds.assign_attrs(renamed_attrs)
            all_attrs.update(renamed_attrs)

            # BUG: still not working
            # update var names (no longer necessary, fixed with qual_name usage up above)
            # renamed_vars = {
            #     name: f"{sub_ds_name}_{name}" for name in list(sub_ds.keys())
            # }
            # sub_ds = sub_ds.rename_vars(renamed_vars)
            sub_dses[sub_ds_name] = sub_ds  # TODO: how was this not necessary before??

        ds_to_merge = [ds]  # , *list(sub_dses.values())]
        ds_to_merge.extend(sub_dses.values())
        ds_merged = xr.merge(ds_to_merge)
        ds_merged = ds_merged.assign_attrs(ds.attrs)
        ds_merged = ds_merged.assign_attrs(all_attrs)
        return ds_merged

    def __call__(
        self, n: int = None, steps: int = None, keep_config: bool = False, **free_refs
    ) -> xr.Dataset:
        """Run the model simulation, allowing specification of any free variables.
        Variables in submodels need to be defined as dictionaries.

        Example:
            >>> dataset = my_model(steps=20, some_free_var=reno.Normal(5, 2), my_submodel=dict(other_free_var=4))

        Args:
            n (int): Number of simulations to run in parallel, leave ``None`` to use the default
                set on the model.
            steps (int): Number of timesteps to run the simulation for, leave ``None`` to use the
                default set on the model.
            keep_config (bool): Whether to keep any changes made via free ref configurations passed
                in for subsequent simulations. The default is to not do this, keeping the original
                model unchanged.
            **free_refs: Definitions for equations or values for any variables or initial conditions
                in the system.
        """
        # store previous config vals and apply any requested config from call params
        previous = self.config()  # noqa: F841
        config = self.config(**free_refs)  # noqa: F841

        if n is None:
            n = self.n
        if steps is None:
            steps = self.steps

        # run the simulation
        self.simulate(n, steps)

        ds = self.dataset()

        # revert config unless explicitly requested to keep it
        if not keep_config:
            self.config(**previous)

        return ds

    def find_timeref_name(self) -> str:
        """If a TimeRef is ever used in an equation, get its reference name. This is
        necessary for the pymc code construction to ensure the correct variable name is
        injected.

        Returns ``None`` if no TimeRef is used.
        """
        all_refs = []
        for flow in self.all_flows():
            # all_refs.extend(flow.eq.seek_refs())
            all_refs.extend(flow.find_refs_of_type(reno.components.TimeRef))
        for var in self.all_vars():
            # all_refs.extend(var.eq.seek_refs())
            all_refs.extend(var.find_refs_of_type(reno.components.TimeRef))

        for ref in all_refs:
            if isinstance(ref, reno.components.TimeRef):
                return ref.name

        return None

    def get_timeref(self) -> reno.components.TimeRef:
        all_refs = []
        for flow in self.all_flows():
            # all_refs.extend(flow.eq.seek_refs())
            all_refs.extend(flow.eq.find_refs_of_type(reno.components.TimeRef))
        for var in self.all_vars():
            # all_refs.extend(var.eq.seek_refs())
            all_refs.extend(var.eq.find_refs_of_type(reno.components.TimeRef))

        for ref in all_refs:
            if isinstance(ref, reno.components.TimeRef):
                return ref

    def dependency_compute_order(
        self, inits_order: bool = False, debug: bool = False
    ) -> list[reno.components.TrackedReference]:
        """Find a dependency-safe ordering for reference equations by iterating through
        and each time adding the first reference that doesn't depend on any references
        not yet added.

        This function will detect circular references (equations that depend on eachother)
        and throw an error. Primary use for this function is to correctly set order of
        equations in the pymc model/step function, but this is used in normal Reno math too.

        Args:
            inits_order (bool): Include stock init equations in the ordering. (When False,
                stocks will always be listed first since they always depend on a previous
                timestep's values. This is not the case for inital equations.)

        Returns:
            The dependency-ordered list of reno TrackedReferences for this model.
        """
        compute_order = []
        # TODO: do we need to ensure _find_all_extended_op_implicit_components
        # is called here?

        if not inits_order:
            # if this isn't an initial case, we assume stocks are always
            # computed first, and this won't cause dependency issues because
            # they're always based in previous values.
            rel_compute_order = reno.utils.dependency_compute_order(
                self.all_flows() + self.all_vars(), init_eqs=False, debug=debug
            )
            compute_order.extend(self.all_stocks())
            compute_order.extend(rel_compute_order)
        else:
            compute_order = reno.utils.dependency_compute_order(
                self.all_stocks() + self.all_flows() + self.all_vars(),
                init_eqs=True,
                debug=debug,
            )

        return compute_order

    def pymc_model(
        self, observations: list["reno.ops.Observation"] = None, steps: int = None
    ) -> pm.model.core.Model:
        """Generate a pymc model for bayesian analysis of this system dynamics model. The general
        idea is that this creates corresponding pymc variables (or distributions as relevant) for
        each stock/flow/var in the model, and sets up the full simulation sequence computations
        based on the generated step function from ``_pt_step()``.

        Sampling with priors should be equivalent to running the system dynamics model normally
        (this is essentially "forward simulation mode".) Add observations to the pymc model
        variables and sample from posterior predictive to run bayesian analysis/determine how
        distributions of any other variables may be affected.
        """
        return reno.pymc.to_pymc_model(self, observations, steps)

    # TODO: wrap in function option (string function would return model)
    # TODO: option to add in necessary imports
    # TODO: can you run black formatting programmatically on a string?
    def pymc_str(
        self, observations: list["reno.ops.Observation"] = None, steps: int = None
    ) -> str:
        """Construct a string of python code to create a pymc model wrapping this system dynamics
        model. Should be a functional (string) equivalent of the ``pymc_model()`` function. Includes
        the output from ``pymc.pt_sim_step_str(self)``.

        Expected imports for the resulting code to run:
            >>> import pytensor
            >>> import pytensor.tensor as pt
            >>> from pytensor.ifelse import ifelse
            >>> import pymc as pm
            >>> import numpy as np
        """
        return reno.pymc.to_pymc_model_str(self, observations, steps)

    def pymc(  # noqa: C901
        self,
        n: int = None,
        steps: int = None,
        sampling_kwargs: dict[str, Any] = None,
        compile_kwargs: dict[str, Any] = None,
        compile_faster: bool = False,
        observations: list["reno.ops.Observation"] = None,
        smc: bool = True,
        trace_prior: az.InferenceData = None,
        compute_prior_only: bool = False,
        keep_config: bool = False,
        **free_refs,
    ) -> az.InferenceData:
        """A PyMC equivalent version of a model's __call__, convert the model to PyMC and run the
        simulation/Bayesian analysis.

        Args:
            n (int): Number of simulations to run in parallel, leave ``None`` to use the default
                set on the model.
            steps (int): Number of timesteps to run the simulation for, leave ``None`` to use the
                default set on the model.
            sampling_kwargs (dict): Arguments to pass to the PyMC sampler. Uses ``sample_smc`` by
                default unless the ``smc`` argument is ``False``, in which case it lets PyMC choose
                sampler (NUTS for continuous variables, some variant of Metropolis for discrete.)
            compile_kwargs (dict): Arguments to pass along to the compiler through PyMC. E.g. ``mode``,
                which, if function compilation is taking forever, can for instance be set to
                ``FAST_COMPILE``.
            compile_faster (bool): For some large/complex models, the PyMC/pytensor compilation step can
                take excessively long as it tries to apply a high level of optimizations. Set this to
                ``True`` to bump down the optimization level by one, which should dramatically speed this
                step up. Alternatively you can have more granular control over this by passing a ``mode``
                key to the ``compile_kwargs`` dictionary parameter, see pytensor's documentation for more
                details: https://pytensor.readthedocs.io/en/latest/tutorial/modes.html
            observations (list[reno.Observation]): Observed values (data/evidence) to use for computing
                posteriors, at least one should be specified if not exclusively running priors.
            smc (bool): Whether to use the sequential monte carlo sampler or not, the default is to
                do so - the regular samplers in PyMC tend not to do well if posterior distributions
                might have multiple peaks, see:
                https://www.pymc.io/projects/examples/en/latest/samplers/SMC2_gaussians.html
            trace_prior [az.InferenceData]: If priors for this model have already been run, pass in
                that arviz inference object here, and the prior xarray dataset will be used in the
                output arviz object from this pymc run.
            compute_prior_only (bool): If set to ``True``, don't run the Bayesian inference, only run
                the priors ("forward simulation mode").
            keep_config (bool): Whether to keep any changes made via free ref configurations passed
                in for subsequent simulations. The default is to not do this, keeping the original
                model unchanged.
            **free_refs: Definitions for equations or values for any variables or initial conditions
                in the system.
        """
        self._reset_type_and_shape_info()

        # TODO: observations, expect dict(ref, sigma, data)
        # store previous config vals
        previous = self.config()  # noqa: F841
        config = self.config(**free_refs)  # noqa: F841

        if n is None:
            n = self.n
        if steps is None:
            steps = self.steps

        if sampling_kwargs is None:
            sampling_kwargs = dict()

        if "cores" not in sampling_kwargs:
            sampling_kwargs["cores"] = 4

        if "draws" not in sampling_kwargs:
            sampling_kwargs["draws"] = math.ceil(n / sampling_kwargs["cores"])

        if compile_kwargs is None:
            compile_kwargs = dict()

        if compile_faster and "mode" not in compile_kwargs:
            compile_kwargs["mode"] = compile.mode.Mode(
                linker="cvm_nogc", optimizer="o3"
            )

        # add any implicit metrics if needed
        implicit_metric_refs = []
        if observations is not None:
            for index, obs in enumerate(observations):
                if not isinstance(obs.ref, reno.components.Reference):
                    metric_obj = reno.components.Metric(obs.ref)
                    setattr(self, f"observation_{index}", metric_obj)
                    obs.ref = metric_obj
                    implicit_metric_refs.append(metric_obj)

        with self.pymc_model(steps=steps) as m:
            # add any observation likelihood variables
            if observations is not None:
                for obs in observations:
                    obs.add_tensors(m)

            # sample!
            sample_func = pm.sample_smc if smc else pm.sample
            if observations is None:
                sample_func = pm.sample_prior_predictive
                sampling_kwargs = dict(draws=n)
            if trace_prior is None:
                trace_prior = pm.sample_prior_predictive(
                    n, compile_kwargs=dict(**compile_kwargs)
                )
                # NOTE: sample_prior_predictive will mutate the passed in
                # dictionary, since I'm using it later, I make a separate copy
                # for the sample_prior_predictive call
            if not compute_prior_only:
                # sample_smc throws a fit with extremely few samples and the
                # exception it raises isn't obviously talking about this (a 0-d
                # iteration error), sometimes a recursion error, so raise our
                # own more informative error
                if observations is not None and n < sampling_kwargs["cores"] * 4:
                    raise Exception(
                        f"{n} is too few samples, run with a higher n parameter, e.g. ``.pymc(n=1000)``"
                    )

                # leaving explicit compile_kwargs out for now because in an
                # older pymc version it wasn't implemented for smc (specifically
                # 5.12.0?) An older version of pymc is sometimes necessary if
                # there's weird stalling issues:
                # https://discourse.pymc.io/t/sample-smc-stalls-at-final-stage/15055/20
                if len(compile_kwargs) > 0:
                    sampling_kwargs["compile_kwargs"] = compile_kwargs
                trace = sample_func(
                    **sampling_kwargs
                )  # , compile_kwargs=compile_kwargs)
                if observations is None:
                    trace.add_groups(posterior=trace.prior)

        if compute_prior_only:
            trace = trace_prior
        elif observations is not None:
            trace.extend(trace_prior)
        elif observations is None:
            del trace.prior
            trace.add_groups(prior=trace_prior.prior)
        self.trace = trace
        self.trace_RVs = [rv.name for rv in m.basic_RVs]

        if not keep_config:
            self.config(**previous)
            for metric in implicit_metric_refs:
                self.metrics.remove(metric)

        return trace

    def to_dict(self, root: bool = True) -> dict:
        """Convert the model into a JSON-serializable dictionary."""

        # TODO: not really any reason to separate _lists since the subsequent
        # dictionary's key lists should be equivalent? (except maybe metrics?)
        data = {
            "name": self.name,
            "doc": self.doc,
            "label": self.label,
            "stocks_list": [stock.name for stock in self.stocks],
            "flows_list": [flow.name for flow in self.flows],
            "vars_list": [var.name for var in self.vars],
            "metrics_list": [
                (metric.name, metric.__class__.__name__) for metric in self.metrics
            ],
            "models_list": [model.name for model in self.models],
            "models": {model.name: model.to_dict(False) for model in self.models},
            "stocks": {stock.name: stock.to_dict() for stock in self.stocks},
            "flows": {flow.name: flow.to_dict() for flow in self.flows},
            "vars": {var.name: var.to_dict() for var in self.vars},
            "metrics": {metric.name: metric.to_dict() for metric in self.metrics},
        }
        if root:
            data["timeref_name"] = self.find_timeref_name()
            data["n"] = self.n
            data["steps"] = self.steps
        return data

    def _add_skeleton(
        self, data: dict
    ) -> dict[str, "reno.components.TrackedReference"]:
        """An inner function for from_dict, it builds out everything structure-wise without
        populating equations and details (necessary because references between equations need
        to exist first.)"""
        building_refs = {}

        for stock_name in data["stocks_list"]:
            setattr(self, stock_name, reno.components.Stock())
            stock = getattr(self, stock_name)
            building_refs[stock.qual_name()] = stock
        for flow_name in data["flows_list"]:
            setattr(self, flow_name, reno.components.Flow())
            flow = getattr(self, flow_name)
            building_refs[flow.qual_name()] = flow
        for var_name in data["vars_list"]:
            setattr(self, var_name, reno.components.Variable())
            var = getattr(self, var_name)
            building_refs[var.qual_name()] = var
        for metric_name, metric_type in data["metrics_list"]:
            if metric_type == "Metric":
                setattr(self, metric_name, reno.components.Metric())
            elif metric_type == "Flag":
                setattr(self, metric_name, reno.components.Flag())
            else:
                raise ValueError(f"Metric type {metric_type} not found?!")
            metric = getattr(self, metric_name)
            building_refs[metric.qual_name()] = metric
        for model_name in data["models_list"]:
            sub_model = Model(
                name=model_name,
                label=data["models"][model_name]["label"],
                doc=data["models"][model_name]["doc"],
            )
            setattr(self, model_name, sub_model)
            sub_refs = sub_model._add_skeleton(data["models"][model_name])
            building_refs.update(sub_refs)

        return building_refs

    def _load_refs(
        self, data: dict, refs: dict[str, "reno.components.TrackedReference"]
    ):
        """Populate all equations and details/submodels etc. Assumes _add_skeleton
        has already been recursively run."""
        for stock_name in data["stocks"]:
            getattr(self, stock_name).from_dict(data["stocks"][stock_name], refs)
        for flow_name in data["flows"]:
            getattr(self, flow_name).from_dict(data["flows"][flow_name], refs)
        for var_name in data["vars"]:
            getattr(self, var_name).from_dict(data["vars"][var_name], refs)
        for metric_name in data["metrics"]:
            getattr(self, metric_name).from_dict(data["metrics"][metric_name], refs)
        for model_name in data["models"]:
            submodel = getattr(self, model_name)
            submodel._load_refs(data["models"][model_name], refs)

    @staticmethod
    def from_dict(data: dict) -> "Model":
        """Deserialize a previously saved model definition dictionary, returns
        new Model instance."""
        m = Model(
            name=data["name"],
            n=data["n"],
            steps=data["steps"],
            label=data["label"],
            doc=data["doc"],
        )
        refs = m._add_skeleton(data)
        if data["timeref_name"] is not None:
            refs[data["timeref_name"]] = reno.components.TimeRef()
        m._load_refs(data, refs)

        return m

    def save(self, path: str):
        """Save model definition at specified location. Stores as a JSON using the
        to_dict() method"""
        data = self.to_dict()
        with open(path, "w") as outfile:
            json.dump(data, outfile, indent=4)

    @staticmethod
    def load(path: str) -> "Model":
        """Initalize model from model definition file (JSON)."""
        with open(path) as infile:
            data = json.load(infile)

        return Model.from_dict(data)



"""Tests for system dynamics functionality."""

import os

import pytest

import reno
from reno.components import Flow, Metric, Piecewise, Scalar, Stock, TimeRef, Variable
from reno.model import Model
from reno.pymc import find_historical_tracked_refs


def test_tub_against_insight_maker():
    """The stock/flow outputs for a system where flows reference stocks etc.
    should match what insight maker outputs, to verify that our n+1, n, n-1
    functionality for various eval'd objects is working as expected."""

    t = TimeRef()

    m = Model("tub", steps=7)
    m.water = Stock()
    m.faucet = Flow(
        Piecewise([m.water + 5, Scalar(0)], [t <= Scalar(4), t > Scalar(4)])
    )
    m.drain = Flow(Scalar(2))

    m.water += m.faucet
    m.water -= m.drain

    ds = m()
    assert ds.water.values.tolist()[0] == [0.0, 3.0, 9.0, 21.0, 45.0, 93.0, 91.0]
    assert ds.faucet.values.tolist()[0] == [5.0, 8.0, 14.0, 26.0, 50.0, 0.0, 0.0]


def test_reassigning_eq_to_ref_warns():
    """Very easy to accidentally define a Flow and then assign its equation to the
    reference itself rather than the ref.eq, so make sure this displays a warning."""

    m = Model("tub", steps=7)
    m.water = Stock()
    m.faucet = Flow()
    m.drain = Flow()

    with pytest.warns(RuntimeWarning):
        m.drain = Scalar(2)


def test_multimodel():
    # TODO: see notebook 11
    pass


def test_copy_model_works(tub_model):
    """Getting a deepcopy of a model shouldn't fail."""
    tub_model.copy("new_name")


def test_historical_tracked_refs():
    """A basic model with several index equations for historical values should
    correctly construct a taps dictionary for the tracked references."""
    t = TimeRef()

    m = Model()
    m.v0 = Variable(Scalar(1))
    m.v1 = Variable(t)
    m.v2 = Variable(m.v1.history(t - 1))
    m.v3 = Variable(m.v1.history(t - 2))
    m.v4 = Variable(m.v2.history(t - m.v0))

    assert find_historical_tracked_refs(m) == {m.v1: [-2, -1], m.v2: [-1]}


def test_multiple_consecutive_runs(tub_model):
    """Running a model multiple times should work the same each time/each
    consecutive run starting from scratch"""
    run1 = tub_model(n=10, faucet_volume=6)
    run2 = tub_model(n=10, faucet_volume=6)

    for variable in run1.keys():
        assert (run1[variable].values == run2[variable].values).all()


def test_multiple_multimodel_consecutive_runs(tub_multimodel):
    """Running a multimodel multiple times should work the same each time/each
    consecutive run starting from scratch"""
    run1 = tub_multimodel(n=10, faucet_volume=6)
    # NOTE: (2025-09-22) Taking these assertions out at least for now because tentatively
    # allowing variable values to be the raw values themselves if no sample
    # dimension required.
    # assert isinstance(tub_multimodel.after_drain.loss_multiplier.value, np.ndarray)
    # assert isinstance(tub_multimodel.after_drain.loss.value, np.ndarray)
    run2 = tub_multimodel(n=10, faucet_volume=6)
    # assert isinstance(tub_multimodel.after_drain.loss_multiplier.value, np.ndarray)
    # assert isinstance(tub_multimodel.after_drain.loss.value, np.ndarray)

    for variable in run1.keys():
        assert (run1[variable].values == run2[variable].values).all()


def test_multimodel_dataset_uses_submodel_prefixes(tub_multimodel):
    """The keys in the dataset from a multimodel run should prefix submodel names
    for each submodel reference."""
    run = tub_multimodel()
    assert "after_drain_intake" in run.keys()
    assert "tub_water_level" in run.keys()


# NOTE: (2025-09-20) Taking this test out at least for now because tentatively
# allowing variable values to be the raw values themselves if no sample
# dimension required.
# def test_consecutive_scalar_collapse():
#     """For unknown reasons, an evaluated scalar variable after a run seems to return a
#     raw float instead of a numpy array with a single value."""
#     m = Model()
#     m.v = Variable(1.0)
#
#     m()
#     assert isinstance(m.v.value, np.ndarray)
#     m()
#     assert isinstance(m.v.value, np.ndarray)


def test_subsequent_pymc_prior_calls(tub_model):
    """Running the same pymc model multiple times should have the same results."""
    run1 = tub_model.pymc(10, compute_prior_only=True)
    run2 = tub_model.pymc(10, compute_prior_only=True)

    for variable in run1.prior.keys():
        assert (run1.prior[variable].values == run2.prior[variable].values).all()


def test_symmetric_config(tub_model):
    """Restoring a previous configuration should _exactly_ (symmetrically) restore
    values. This is most obvious in what gets passed to pymc models."""
    previous = tub_model.config()
    assert tub_model.water_level.init is None
    tub_model.config(**previous)
    assert tub_model.water_level.init is None


def test_symmetric_config_pymc_str(tub_model):
    """Restoring a previous configuration should _exactly_ (symmetrically) restore
    values. This is most obvious in what gets passed to pymc models."""
    previous = tub_model.config()
    str1 = tub_model.pymc_str()
    tub_model.config(**previous)
    str2 = tub_model.pymc_str()

    assert str1 == str2


def test_to_from_dict_consistency(tub_multimodel):
    """The serialized from of a deserialized serialized model should
    be the same as the original serialization."""
    serialized1 = tub_multimodel.to_dict()

    new_model = Model.from_dict(serialized1)
    serialized2 = new_model.to_dict()

    assert serialized1 == serialized2


def test_from_dict_result_consistency(tub_multimodel):
    """Loading a previously saved model should have the same results when run."""

    serialized = tub_multimodel.to_dict()
    new_model = Model.from_dict(serialized)

    run1 = tub_multimodel(n=10, faucet_volume=6)
    run2 = new_model(n=10, faucet_volume=6)

    for variable in run1.keys():
        assert (run1[variable].values == run2[variable].values).all()


def test_save_load(data_file_loc, tub_multimodel):
    """Saving to and loading from a file should have the same results when run."""
    run1 = tub_multimodel(n=10, faucet_volume=6)
    filename = f"{data_file_loc}/testmodel.json"

    tub_multimodel.save(filename)
    assert os.path.exists(filename)

    new_model = Model.load(filename)
    run2 = new_model(n=10, faucet_volume=6)

    for variable in run1.keys():
        assert (run1[variable].values == run2[variable].values).all()


def test_multidim_model_runs(multidim_model):
    """A model with stocks and flows with a vector dimension should run without errors
    and all outputs in dataset should have correct shape."""

    ds = multidim_model()
    assert len(ds.coords) == 6
    assert "s_vec" in ds.coords
    assert "v1_vec" in ds.coords
    assert "v3_vec" in ds.coords
    assert "v4_vec" in ds.coords

    assert ds.s.shape == (1, 10, 4)
    assert ds.v1.shape == (1, 4)
    assert ds.v2.shape == (1,)
    assert ds.v3.shape == (1, 10, 4)
    assert ds.v4.shape == (1, 6)


def test_multidim_model_w_vectors_runs(multidim_model_determ):
    """A multidim model with a static vector should still run correctly and have correct dims in the dataset output."""
    ds = multidim_model_determ()
    assert (ds.v1.values == [[0.25, 0.35, 0.45, 0.55]]).all()
    assert ds.v1.shape == (1, 4)
    assert ds.v2.shape == (1,)
    assert ds.v3.shape == (1, 10, 4)
    assert ds.s.shape == (1, 10, 4)


def test_model_context_manager_finds_name():
    """Creating references within a model's context manager should appropriately find the variable names and add them
    to the model"""

    m = Model()
    with m:
        var1 = Variable(2)
        var2 = Variable(5)
        combination = Variable(var1 + var2)
        some_metric = Metric(var1.timeseries[-1])

    assert var1.name == "var1"
    assert var2.name == "var2"
    assert combination.name == "combination"
    assert some_metric.name == "some_metric"
    assert var1.model == m
    assert var2.model == m
    assert combination.model == m
    assert some_metric.model == m

    assert m.var1 == var1
    assert m.var2 == var2
    assert m.combination == combination
    assert m.some_metric == some_metric

    ds = m(1, 1)
    assert ds.combination.values[0] == 7


def test_model_context_manager_ignores_preexisting():
    """Creating references within a model's context manager should appropriately find the variable names and add them
    to the model unless it's explicitly added to the model already"""

    m = Model()
    with m:
        var1 = Variable(2)
        m.var2 = Variable(5)
        combination = Variable(var1 + m.var2)

    assert var1.name == "var1"
    assert m.var2.name == "var2"
    assert combination.name == "combination"
    assert var1.model == m
    assert m.var2.model == m
    assert combination.model == m

    assert m.var1 == var1
    assert m.combination == combination

    ds = m(1, 1)
    assert ds.combination.values[0] == 7


def test_model_context_manager_handles_submodels():
    """Creating models within a model's context manager should appropriately find the model name and add to parent
    model."""

    m1 = Model()
    with m1:
        var1 = Variable(2)
        m2 = Model()
        with m2:
            var2 = Variable(5)
        m1.combination = Variable(var1 + m2.var2)

    assert m2.parent == m1
    assert m1.m2 == m2
    assert m1.m2.var2 == var2
    assert m2 in m1.models
    assert m2.name == "m2"

    ds = m1(1, 1)
    assert ds.combination.values[0] == 7


def test_bottleneck_example():
    """A model using the .outflows or .space of a stock to help limit should correctly bottleneck upstream."""
    m = Model()
    with m:
        f_in, f_between, f_out = Flow(), Flow(), Flow(1)
        s0, s1, sfinal = Stock(max=2), Stock(max=3), Stock()

        f_in.eq = reno.minimum(4, s0.space)
        f_between.eq = reno.minimum(2, s1.space)

        s0 += f_in
        s0 -= f_between
        s1 += f_between
        s1 -= f_out
        sfinal += f_out

    ds = m()
    assert (ds.f_in[0][0:4] == [4.0, 2.0, 2.0, 1.0]).all()
    assert (ds.f_between[0][0:4] == [2.0, 2.0, 2.0, 1.0]).all()


def test_bottleneck_example_pymc():
    """A pymc model using the .outflows or .space of a stock to help limit should correctly bottleneck upstream."""
    m = Model()
    with m:
        f_in, f_between, f_out = Flow(), Flow(), Flow(1)
        s0, s1, sfinal = Stock(max=2), Stock(max=3), Stock()

        f_in.eq = reno.minimum(4, s0.space)
        f_between.eq = reno.minimum(2, s1.space)

        s0 += f_in
        s0 -= f_between
        s1 += f_between
        s1 -= f_out
        sfinal += f_out

    ds1 = m()
    ds2 = m.pymc(n=1, compute_prior_only=True)
    assert (ds1.f_in.values == ds2.prior.f_in.sel(chain=0).values).all()
    assert (ds1.f_between.values == ds2.prior.f_between.sel(chain=0).values).all()
    assert (ds1.s0.values == ds2.prior.s0.sel(chain=0).values).all()
    assert (ds1.s1.values == ds2.prior.s1.sel(chain=0).values).all()



"""Classes for math operations, these are used to build up symbolic equations,
similar in principle to what something like PyTensor is doing."""

import math
import warnings

import numpy as np
import pymc as pm
import pytensor.tensor as pt

import reno

__all__ = [
    # -- series math operations --
    "series_max",
    "series_min",
    "mean",
    "sum",
    "nonzero",
    "diff",
    "nanindex",
    "index",
    "slice",
    "orient_timeseries",
    # -- normal math operations --
    "add",
    "sub",
    "mul",
    "div",
    "mod",
    "abs",
    "lt",
    "lte",
    "gt",
    "gte",
    "eq",
    "ne",
    "bool_and",
    "bool_or",
    "minimum",
    "maximum",
    "clip",
    "log",
    "sin",
    "interpolate",
    "assign",
    "astype",
    "stack",
    # -- "higher order" --
    "pulse",
    "repeated_pulse",
    "step",
    "delay1",
    "delay3",
    "smooth",
    # -- meta --
    "inflow",
    "outflows",
    # -- distributions --
    "Normal",
    "Uniform",
    "DiscreteUniform",
    "Bernoulli",
    "Categorical",
    "List",
    "Observation",
    # "Sweep",
]


# ==================================================
# SERIES MATH OPERATIONS
# ==================================================
# These are primarily meant for metric equations that run after a
# simulation and operate on a whole timeseries at a time, rather
# than calculating for a single timestep


# NOTE: can't use 'proper' name of max because TrackedReferences already have a
# max (equation max) which I don't want to rename.
class series_max(reno.components.Operation):
    """Maximum value in an array. Effectively a row-wise np.max. This can either be applied
    to a timeseries or a value with a data dimension.

    Example:
        .. code-block:: python

            >>> import reno as r
            >>> a = r.Variable([6, 5, 4, 8])
            >>> a.series_max().eval()
            np.int64(8)

            >>> m = r.Model()
            >>> t = r.TimeRef()
            >>> m.v0 = r.Variable(t + 5)
            >>> m.max_value = r.Metric(m.v0.timeseries.series_max())
            >>> m()
            >>> m.v0.value
            array([[ 5,  6,  7,  8,  9, 10, 11, 12, 13, 14]])

            >>> m.max_value.value
            array([14])
    """

    OP_REPR = "max"

    def __init__(self, a):
        super().__init__(a)

    def latex(self, **kwargs):
        return f"\\text{{max}}({self.sub_equation_parts[0].latex(**kwargs)})"

    def get_shape(self) -> int:
        return 1

    def op_eval(self, **kwargs):
        value = self.sub_equation_parts[0].value
        if value is None:
            value = self.sub_equation_parts[0].eval(**kwargs)
        axis = 0
        if len(value.shape) != 1:
            axis += 1
        return np.nanmax(value, axis=axis)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.max(self.sub_equation_parts[0].pt(**refs))

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.max({self.sub_equation_parts[0].pt_str(**refs)})"


# NOTE: can't use 'proper' name of min because TrackedReferences already have a
# min (equation min) which I don't want to rename.
class series_min(reno.components.Operation):
    """Minimum value in an array. Effectively a row-wise np.min. This can either be applied
    to a timeseries or a value with a data dimension.

    Example:
        .. code-block:: python

            >>> import reno as r
            >>> a = r.Variable([6, 5, 4, 8])
            >>> a.series_min().eval()
            np.int64(4)

            >>> m = r.Model()
            >>> t = r.TimeRef()
            >>> m.v0 = r.Variable(t + 5)
            >>> m.min_value = r.Metric(m.v0.timeseries.series_min())
            >>> m()
            >>> m.v0.value
            array([[ 5,  6,  7,  8,  9, 10, 11, 12, 13, 14]])

            >>> m.min_value.value
            array([5])
    """

    OP_REPR = "min"

    def __init__(self, a):
        super().__init__(a)

    def latex(self, **kwargs):
        return f"\\text{{min}}({self.sub_equation_parts[0].latex(**kwargs)})"

    def get_shape(self) -> int:
        return 1

    def op_eval(self, **kwargs):
        value = self.sub_equation_parts[0].value
        if value is None:
            value = self.sub_equation_parts[0].eval(**kwargs)
        axis = 0
        if len(value.shape) != 1:
            axis += 1
        return np.nanmin(value, axis=axis)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.min(self.sub_equation_parts[0].pt(**refs))

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.min({self.sub_equation_parts[0].pt_str(**refs)})"


class mean(reno.components.Operation):
    """Average across the values of a vector (either timeseries or data dim.)

    Example:
        .. code-block:: python

            >>> import reno as r
            >>> a = r.Variable([0, 1, 2, 3])
            >>> a.mean().eval()
            np.float64(1.5)

            >>> m = r.Model()
            >>> t = r.TimeRef()
            >>> m.v0 = r.Variable(t + 1)
            >>> m.v1 = r.Variable(m.v0.timeseries[t-2:t].mean())
            >>> m()
            >>> m.v0.value
            array([[ 1,  2,  3,  4,  5,  6,  7,  8,  9, 10]])

            >>> m.v1.value
            array([[nan, 1. , 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5]])
    """

    def __init__(self, a, axis=0):
        self.axis = axis
        super().__init__(a)

    def latex(self, **kwargs):
        return f"\\text{{mean}}({self.sub_equation_parts[0].latex(**kwargs)})"

    def get_shape(self) -> int:
        return 1

    def get_type(self) -> type:
        return float

    def op_eval(self, **kwargs):
        # value = self.sub_equation_parts[0].value
        # if value is None:
        value = self.sub_equation_parts[0].eval(**kwargs)
        axis = self.axis
        if len(value.shape) != 1:
            axis = self.axis + 1  # to account for "batch"/n dimension
        return np.nanmean(value, axis=axis)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.mean(self.sub_equation_parts[0].pt(**refs), axis=self.axis)

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.mean({self.sub_equation_parts[0].pt_str(**refs)}, axis={self.axis})"


class sum(reno.components.Operation):
    """Series-wise sum (e.g. row-wise if a matrix)."""

    # NOTE: sums of static values won't give you a multiple by timesteps by default, e.g.
    # if you have a static variable with value 5, the variable.sum() will return 5. In order
    # to expand a static value based on the timeseries, access a slice of the variable. For
    # instance, to get the full series of static values, use variable[:].sum()

    # NOTE: NOTE: nope, since changed to make operand automatically a slice if it's static.
    # Remember that this operation is really only meant for metrics, use history components
    # for equations within the stocks/flows

    def __init__(self, a, axis=0):
        self.axis = axis
        super().__init__(a)
        # if a.is_static() and not isinstance(a, reno.ops.time_slice):
        #     super().__init__(reno.ops.time_slice(a))
        # else:
        #     super().__init__(a)

    def latex(self, **kwargs):
        return f"\\Sigma {self.sub_equation_parts[0].latex(**kwargs)}"

    def get_shape(self) -> int:
        return 1

    def op_eval(self, **kwargs):
        # value = self.sub_equation_parts[0].value
        # if value is None:
        value = self.sub_equation_parts[0].eval(**kwargs)
        axis = self.axis
        if len(value.shape) != 1:
            axis = self.axis + 1  # to account for "batch"/n dimension
        return np.nansum(value, axis=axis)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.sum(self.sub_equation_parts[0].pt(**refs), axis=self.axis)

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.sum({self.sub_equation_parts[0].pt_str(**refs)}, axis={self.axis})"


class nonzero(reno.components.Operation):
    """Similar to the numpy nonzero or the numpy ``where`` specified with only a condition.

    Note that in numpy evaluation this returns a "jagged" array, padded at the end by np.NaN values.
    Also note that since np.NaN is a float value, the resulting array is converted to floats.

    Example:
        .. code-block::

            >>> import reno as r
            >>> v = r.Variable([0, 0, 1, 0, 0, 0, 1])
            >>> r.nonzero(v).eval()
            array([2., 6., nan, nan, nan, nan, nan])
    """

    def __init__(self, a):
        super().__init__(a)

    def get_type(self) -> type:
        return int

    def op_eval(self, **kwargs):

        value = self.sub_equation_parts[0].eval(**kwargs)
        indices = np.nonzero(value)

        # say we start with
        # np.array([
        #   [0, 0, 1, 0, 0, 0, 1],
        #   [1, 0, 0, 1, 0, 0, 1]
        # ])

        # the np.nonzero returns:
        # (array([0, 0, 1, 1, 1]), array([2, 6, 0, 3, 6]))

        # the goal is to have a "jagged" matrix with the indices (2, 6, 0, 3, 6)
        # as the left-most columns and of course in the appropriate rows

        # create full matrix of nan's as padding
        result = np.empty(value.shape)
        result[:] = np.nan

        if len(value.shape) == 2:
            # create incrementing sequences for the columns for each row index
            conditions = [indices[0] == val for val in np.unique(indices[0])]
            # e.g.
            # [array([ True,  True, False, False, False]),
            #  array([False, False,  True,  True,  True])]
            # make the sequences
            ranges = [np.arange(0, condition.sum()) for condition in conditions]
            # e.g. [array([0, 1]), array([0, 1, 2])]
            # np.select requires a set of matching size arrays, so we make a bunch
            # of zeros with the sequences embedded where the conditions dictate
            condition_ranges = [np.zeros(condition.shape) for condition in conditions]
            for i in range(len(condition_ranges)):
                condition_ranges[i][conditions[i]] = ranges[i]
            # e.g. [array([0., 1., 0., 0., 0.]), array([0., 0., 0., 1., 2.])]
            index_indexers = np.select(conditions, condition_ranges).astype(int)
            # e.g. array([0., 1., 0., 1., 2.])
            # this fills the first part of every row with the column indices from
            # the initial nonzero call
            # so equivalent to result[[0, 0, 1, 1, 1], [0, 1, 0, 1, 2]] = [2, 6, 0, 3, 6]
            result[indices[0], index_indexers] = indices[1]
        elif len(value.shape) == 1:
            print(indices[0])
            result[: len(indices[0])] = indices[0]
        else:
            raise RuntimeError("Cannot yet run nonzero op on an extra data dimension")

        # so going from:
        # np.array([
        #   [0, 0, 1, 0, 0, 0, 1],
        #   [1, 0, 0, 1, 0, 0, 1]
        # ])
        # to:
        # array([
        #   [ 2.,  6., nan, nan, nan, nan, nan],
        #   [ 0.,  3.,  6., nan, nan, nan, nan]
        # ]

        return result

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        # TODO: I don't think this handles extra data dim?
        return pt.nonzero(self.sub_equation_parts[0].pt(**refs))[0]

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.nonzero({self.sub_equation_parts[0].pt_str(**refs)})[0]"


class diff(reno.components.Operation):
    """Returns the difference between every value in the list and the value before it. Note that the size of the output list is one less than the input, so likely need to process in some other way instead of directly including in e.g. a metric."""

    def __init__(self, a):
        super().__init__(a)

    def op_eval(self, **kwargs):
        value = self.sub_equation_parts[0].eval(**kwargs)
        return np.diff(value)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.diff(self.sub_equation_parts[0].pt(**refs))

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.diff({self.sub_equation_parts[0].pt_str(**refs)})"


class nanindex(reno.components.Operation):
    """Special index op for jagged arrays that come from the nonzero op. These convert any negative indices to
    ignore any nans at the end of each row. Is equivalent to normal index for pymc math.
    """

    def __init__(self, a, ind):
        super().__init__(a, ind)

    def latex(self, **kwargs) -> str:
        return f"{self.sub_equation_parts[0].latex(**kwargs)}[{self.sub_equation_parts[1].latex(**kwargs)}]"

    def get_shape(self) -> int:
        return 1

    def op_eval(self, **kwargs):
        # TODO: support for static?
        value = self.sub_equation_parts[0].eval(**kwargs)
        indices = self.sub_equation_parts[1].eval(**kwargs)
        # should axis that's 1 actually be -1?
        indices = np.where(
            indices < 0, (~np.isnan(value)).sum(axis=-1) + indices, indices
        )
        # return value[..., indices]
        return np.choose(indices, value.T)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs)[
            self.sub_equation_parts[1].pt(**refs)
        ]

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"{self.sub_equation_parts[0].pt_str(**refs)}[{self.sub_equation_parts[1].pt_str(**refs)}]"


class index(reno.components.Operation):
    """Get a previous value in the time series at specified index, only works for tracked references
    inside of equations for metrics."""

    def __init__(self, a, ind):
        super().__init__(a, ind)

    def latex(self, **kwargs) -> str:
        return f"{self.sub_equation_parts[0].latex(**kwargs)}[{self.sub_equation_parts[1].latex(**kwargs)}]"

    def get_shape(self) -> int:
        return 1

    def op_eval(self, **kwargs):
        # TODO: support for static?
        value = self.sub_equation_parts[0].eval(**kwargs)
        # if len(value.shape) > 1:
        #     return value[:, self.sub_equation_parts[1].value]  # TODO: eval sub_eq_parts[1]?
        # else:
        #     return value[self.sub_equation_parts[1].value]  # TODO: eval sub_eq_parts[1]?

        # return value[..., self.sub_equation_parts[1].value]
        # TODO: does this need to use np.choose instead?
        indices = self.sub_equation_parts[1].eval(**kwargs)
        if isinstance(indices, np.ndarray) and len(indices) > 1:
            return np.choose(indices, value.T)
        else:
            return value[..., indices]

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs)[
            self.sub_equation_parts[1].pt(**refs)
        ]

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"{self.sub_equation_parts[0].pt_str(**refs)}[{self.sub_equation_parts[1].pt_str(**refs)}]"


class slice(reno.components.Operation):
    """Can be applied along with timeseries op in metrics for getting specific time segments, or
    can be applied generally in equations when dealing with vector data.

    Similar to normal python slices, top is exclusive
    """

    def __init__(self, a, start=None, stop=None):
        self.start = start
        self.stop = stop

        if self.start is not None:
            # keep these ensure_scalars since start/stop stored
            # on this class, and the ensure_scalar applied to sub_equation_parts
            # in EquationPart obv won't magically apply to these
            self.start = reno.utils.ensure_scalar(self.start)

        if self.stop is not None:
            self.stop = reno.utils.ensure_scalar(self.stop)

        operands = [a]
        if start is not None:
            operands.append(self.start)
        if stop is not None:
            operands.append(self.stop)
        super().__init__(*operands)

    def latex(self, **kwargs) -> str:
        start = "" if self.start is None else self.start.latex(**kwargs)
        stop = "" if self.stop is None else self.stop.latex(**kwargs)
        return f"{self.sub_equation_parts[0].latex(**kwargs)}[{start}:{stop}]"

    # TODO: (2025.09.22) will need to think about how best to implement a
    # get_shape for this

    def op_eval(self, t, **kwargs):
        start = self.start.eval(t, **kwargs) if self.start is not None else None
        stop = self.stop.eval(t, **kwargs) if self.stop is not None else None

        if isinstance(start, np.ndarray):
            start = start[0]
        if isinstance(stop, np.ndarray):
            stop = stop[0]

        if start is None or start < 0:
            start = 0
        if stop is not None and stop < start:
            stop = start

        value = self.sub_equation_parts[0].value
        if value is None:
            value = self.sub_equation_parts[0].eval(t, **kwargs)

        # dims = []
        # Note we don't have explicit handling for statics, use of .timeseries
        # op is assumed where necessary.
        # TODO: is it cleaner to use ellipsis? https://stackoverflow.com/questions/12116830/numpy-slice-of-arbitrary-dimensions
        # for i, dim in enumerate(value.shape):
        #     if i == len(value.shape) - 1:
        #         dims.append(slice(start, stop))
        #     else:
        #         dims.append(slice(None, None))
        # print(dims)
        return value[..., start:stop]

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        if self.start is None and self.stop is None:
            return self.sub_equation_parts[0].pt(**refs)
        t_0 = (
            pt.as_tensor(0)
            if self.start is None
            else pt.maximum(pt.as_tensor(0), self.start.pt(**refs))
            # else self.start.pt(**refs)
        )
        if self.stop is None:
            return self.sub_equation_parts[0].pt(**refs)[..., t_0:]
        return self.sub_equation_parts[0].pt(**refs)[
            ..., t_0 : pt.maximum(self.stop.pt(**refs), t_0)
        ]

    def pt_str(self, **refs: dict[str, str]) -> str:
        if self.start is None and self.stop is None:
            return f"{self.sub_equation_parts[0].pt_str(**refs)}"
        t0 = (
            "pt.as_tensor(0)"
            if self.start is None
            else f"pt.maximum(pt.as_tensor(0), {self.start.pt_str(**refs)})"
        )
        if self.stop is None:
            return f"{self.sub_equation_parts[0].pt_str(**refs)}[..., {t0}:]"
        return f"{self.sub_equation_parts[0].pt_str(**refs)}[..., {t0} : pt.maximum({self.stop.pt_str(**refs)}, {t0})]"


class orient_timeseries(reno.components.Operation):
    """Get the full data of a component/equation as an array over time, allowing aggregate
    operations to operate across the full timeseries, e.g. for metric equations. This is
    in essence "reorienting" the underlying data to include the time dimension.

    NOTE: for pytensor I suspect this will only work in metric equations. pt/pt_str are dependent
    on whether the refs are populated with the full series or not, which will only occur
    within the metrics section?
    TODO: we can _eventually_ make this work within pytensor by the approach to solving general
    history time equations - if a timeseries op detected within non-metric context, include the full
    stops of the relevant variables

    NOTE: only use this in a non-metric equation if a _range_ of timeseries values are needed. Otherwise use `.history()` for single values
    """

    # TODO: it probably doesn't make sense to call this on anything except a trackedreference,
    # may want to add checks for this.

    def __init__(self, a):
        super().__init__(a)

    def latex(self, **kwargs) -> str:
        return f"{self.sub_equation_parts[0].latex(**kwargs)}.\\text{{timeseries}}"

    def __repr__(self):
        """Explicit repr is necessary to avoid including a timeref in sub equation parts if added from
        the pt() or pt_str() calls."""
        return f"({self.op_repr()} {self.sub_equation_parts[0].__repr__()})"

    # TODO: how best to get timeseries length for get_shape?

    def op_eval(self, t, **kwargs):
        value = self.sub_equation_parts[0].value
        if value is None:
            value = self.sub_equation_parts[0].eval(
                t, **kwargs
            )  # TODO: is this right?? This isn't the same as getting .value because may not include full time series...
            # do I instead need to ensure the save is true and then get .value?
            # this will only be relevant to solve when the pytensor general
            # history approach is solved

        count = t + 1
        if self.sub_equation_parts[0].is_static():
            remaining = 0
            if hasattr(self.sub_equation_parts[0], "model"):
                if self.sub_equation_parts[0].model is not None:
                    remaining = self.sub_equation_parts[0].model.steps - count
            if isinstance(value, (float, int)):
                return np.concatenate(
                    [
                        np.array([np.array([value]).repeat(count)]),
                        np.array([np.array([0]).repeat(remaining)]),
                    ],
                    axis=1,
                )
            if len(value.shape) == 1:
                return value[:, None].repeat(count, axis=1)
        return value

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        # NOTE: have to have a time ref for this to work, so we add one if none
        # found. As a result, have to make sure
        if self.sub_equation_parts[0].model.find_timeref_name() is None:
            self.sub_equation_parts.append(reno.components.TimeRef())

        # for static underlying values, we reconstruct an array as if that
        # underlying value were building up through an array of all zeros,
        # so ""fill"" accordingly based on timestep
        if self.sub_equation_parts[0].is_static():
            seq_length = refs["__PT_SEQ_LEN__"]
            if "__TIMESERIES__" in refs:
                if refs["__TIMESERIES__"] == "init":
                    return pt.concatenate(
                        [
                            [self.sub_equation_parts[0].pt(**refs)],
                            pt.as_tensor(0).repeat(seq_length - 1),
                        ]
                    )
                elif refs["__TIMESERIES__"] == "scan":
                    t = refs[self.sub_equation_parts[0].model.find_timeref_name()]
                    return pt.concatenate(
                        [
                            self.sub_equation_parts[0].pt(**refs).repeat(t + 1),
                            pt.as_tensor(0).repeat(seq_length - (t + 1)),
                        ]
                    )
            # in principle this should never be reached?
            return pt.repeat(self.sub_equation_parts[0].pt(**refs), seq_length)
        if "__TIMESERIES__" in refs:
            name = self.sub_equation_parts[0].qual_name()
            # NOTE: _h is what's stored in refs during init and scan,
            # assuming I still need this to ensure the post calculation
            # is done correctly with the full sequence instead of the historical
            # ones.
            if refs["__TIMESERIES__"] == "init":
                name += "_h"
                # TODO: this doesn't work if no timeref included.
                # t = refs[self.sub_equation_parts[0].model.find_timeref_name()]
                # have to concatenate to reorder, since pymc is passing in
                # history (taps) according to the current timestep.
                return pt.concatenate([refs[name][-1:], refs[name][:-1]])
            elif refs["__TIMESERIES__"] == "scan":
                name += "_h"
                # TODO: this doesn't work if no timeref included.
                t = refs[self.sub_equation_parts[0].model.find_timeref_name()]
                # have to concatenate to reorder, since pymc is passing in
                # history (taps) according to the current timestep.
                return pt.concatenate([refs[name][-(t + 1) :], refs[name][: -(t + 1)]])
            return refs[name]
        return self.sub_equation_parts[0].pt(**refs)

    def pt_str(self, **refs: dict[str, str]) -> str:
        # have to have a time ref for this to work
        if self.sub_equation_parts[0].model.find_timeref_name() is None:
            self.sub_equation_parts.append(reno.components.TimeRef())

        # for static underlying values, we reconstruct an array as if that
        # underlying value were building up through an array of all zeros,
        # so ""fill"" accordingly based on timestep
        if self.sub_equation_parts[0].is_static():
            seq_length = refs["__PT_SEQ_LEN__"]
            if "__TIMESERIES__" in refs:
                if refs["__TIMESERIES__"] == "init":
                    return f"pt.concatenate([[{self.sub_equation_parts[0].pt_str(**refs)}], pt.as_tensor(0).repeat({seq_length - 1})])"
                elif refs["__TIMESERIES__"] == "scan":
                    t = refs[self.sub_equation_parts[0].model.find_timeref_name()]
                    return f"pt.concatenate([{self.sub_equation_parts[0].pt_str(**refs)}.repeat(t + 1), pt.as_tensor(0).repeat({seq_length} - ({t} + 1))])"
            # in principle this should never be reached?
            return (
                f"pt.repeat({self.sub_equation_parts[0].pt_str(**refs)}, {seq_length})"
            )
        if "__TIMESERIES__" in refs:
            name = self.sub_equation_parts[0].qual_name()
            if refs["__TIMESERIES__"] == "init":
                name += "_h"
                # t = refs[self.sub_equation_parts[0].model.find_timeref_name()]
                return f"pt.concatenate([{refs[name]}[-1:], {refs[name]}[:-1]])"
            elif refs["__TIMESERIES__"] == "scan":
                name += "_h"
                t = refs[self.sub_equation_parts[0].model.find_timeref_name()]
                return f"pt.concatenate([{refs[name]}[-({t}+1):], {refs[name]}[:-({t}+1)]])"
            return refs[name]
        else:
            return self.sub_equation_parts[0].pt_str(**refs)


# ==================================================
# NORMAL MATH OPERATIONS
# ==================================================


def adjust_shapes_for_n(*parts, **kwargs):
    """for numpy calculations, while () can broadcast to (dim,)
    but not (n,) to (n,dim), so handle converting (n,) to (n,1)
    if necessary, given the other components.

    Note that this runs and returns the evaluations for each part.
    """
    dims = [part.shape for part in parts]
    vals = [part.eval(**kwargs) for part in parts]
    if max(dims) == 1:
        return vals
    for i, val in enumerate(vals):
        if dims[i] == 1:
            vals[i] = np.expand_dims(val, -1)
    return vals


class add(reno.components.Operation):
    """a + b

    Example:
        .. code-block:: python

            >>> import reno as r
            >>> r.Scalar(3) + 2
            (+ Scalar(3) Scalar(2))

            >>> (r.Scalar(3) + 2).eval()
            5

            >>> a = r.Variable([0, 1, 2])
            >>> b = r.Variable([3, 4, 5])
            >>> (a + b).eval()
            array([3, 5, 7])
    """

    OP_REPR = "+"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} + {self.sub_equation_parts[1].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        a, b = adjust_shapes_for_n(*self.sub_equation_parts[0:2], **kwargs)
        return a + b

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) + self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} + {self.sub_equation_parts[1].pt_str(**refs)})"


class sub(reno.components.Operation):
    """a - b

    Example:
        .. code-block:: python

            >>> import reno as r
            >>> r.Scalar(3) - 2
            (- Scalar(3) Scalar(2))

            >>> (r.Scalar(3) - 2).eval()
            1

            >>> a = r.Variable([0, 1, 2])
            >>> b = r.Variable([3, 4, 5])
            >>> (b - a).eval()
            array([3, 3, 3])
    """

    OP_REPR = "-"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} - {self.sub_equation_parts[1].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        a, b = adjust_shapes_for_n(*self.sub_equation_parts[0:2], **kwargs)
        return a - b

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) - self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} - {self.sub_equation_parts[1].pt_str(**refs)})"


class mul(reno.components.Operation):
    """a * b

    Example:
        .. code-block:: python

            >>> import reno as r
            >>> r.Scalar(3) * 2
            (* Scalar(3) Scalar(2))

            >>> (r.Scalar(3) * 2).eval()
            6

            >>> a = r.Variable([0, 1, 2])
            >>> b = r.Variable([3, 4, 5])
            >>> (a * b).eval()
            array([0, 4, 10])
    """

    OP_REPR = "*"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} * {self.sub_equation_parts[1].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        a, b = adjust_shapes_for_n(*self.sub_equation_parts[0:2], **kwargs)
        return a * b

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) * self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} * {self.sub_equation_parts[1].pt_str(**refs)})"


class div(reno.components.Operation):
    """a / b

    Example:
        .. code-block:: python

            >>> import reno as r
            >>> r.Scalar(4) / 2
            (/ Scalar(4) Scalar(2))

            >>> (r.Scalar(4) / 2).eval()
            2.0

            >>> a = r.Variable([0, 1, 2])
            >>> b = r.Variable([3, 4, 5])
            >>> (a / b).eval()
            array([0, 0.25, 0.4])
    """

    OP_REPR = "/"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"\\frac{{{self.sub_equation_parts[0].latex(**kwargs)}}}{{{self.sub_equation_parts[1].latex(**kwargs)}}}"

    def op_eval(self, **kwargs):
        a, b = adjust_shapes_for_n(*self.sub_equation_parts[0:2], **kwargs)
        return a / b

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) / self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} / {self.sub_equation_parts[1].pt_str(**refs)})"


class mod(reno.components.Operation):
    """a % b"""

    OP_REPR = "%"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} \\% {self.sub_equation_parts[1].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        a, b = adjust_shapes_for_n(*self.sub_equation_parts[0:2], **kwargs)
        return a % b

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.mod(
            self.sub_equation_parts[0].pt(**refs), self.sub_equation_parts[1].pt(**refs)
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.mod({self.sub_equation_parts[0].pt_str(**refs)}, {self.sub_equation_parts[1].pt_str(**refs)})"


class abs(reno.components.Operation):
    """|a| (absolute value)"""

    def __init__(self, a):
        super().__init__("abs", a)

    def latex(self, **kwargs):
        return f"|{self.sub_equation_parts[0].latex(**kwargs)}|"

    def op_eval(self, **kwargs):
        return np.abs(self.sub_equation_parts[0].eval(**kwargs))

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.abs(self.sub_equation_parts[0].pt(**refs))

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.abs({self.sub_equation_parts[0].pt_str(**refs)})"


class lt(reno.components.Operation):
    """a < b"""

    OP_REPR = "<"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} < {self.sub_equation_parts[1].latex(**kwargs)}"

    def get_type(self) -> type:
        return bool

    def op_eval(self, **kwargs):
        a, b = adjust_shapes_for_n(*self.sub_equation_parts[0:2], **kwargs)
        return a < b

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) < self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} < {self.sub_equation_parts[1].pt_str(**refs)})"


class lte(reno.components.Operation):
    """a <= b"""

    OP_REPR = "<="

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} \\leq {self.sub_equation_parts[1].latex(**kwargs)}"

    def get_type(self) -> type:
        return bool

    def op_eval(self, **kwargs):
        a, b = adjust_shapes_for_n(*self.sub_equation_parts[0:2], **kwargs)
        return a <= b

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) <= self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} <= {self.sub_equation_parts[1].pt_str(**refs)})"


class gt(reno.components.Operation):
    """a > b"""

    OP_REPR = ">"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} > {self.sub_equation_parts[1].latex(**kwargs)}"

    def get_type(self) -> type:
        return bool

    def op_eval(self, **kwargs):
        a, b = adjust_shapes_for_n(*self.sub_equation_parts[0:2], **kwargs)
        return a > b

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) > self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} > {self.sub_equation_parts[1].pt_str(**refs)})"


class gte(reno.components.Operation):
    """a >= b"""

    OP_REPR = ">="

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} \\geq {self.sub_equation_parts[1].latex(**kwargs)}"

    def get_type(self) -> type:
        return bool

    def op_eval(self, **kwargs):
        a, b = adjust_shapes_for_n(*self.sub_equation_parts[0:2], **kwargs)
        return a >= b

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) >= self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} >= {self.sub_equation_parts[1].pt_str(**refs)})"


class eq(reno.components.Operation):
    """a == b"""

    OP_REPR = "=="

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} = {self.sub_equation_parts[1].latex(**kwargs)}"

    def get_type(self) -> type:
        return bool

    def op_eval(self, **kwargs):
        a, b = adjust_shapes_for_n(*self.sub_equation_parts[0:2], **kwargs)
        return a == b

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.eq(
            self.sub_equation_parts[0].pt(**refs),
            self.sub_equation_parts[1].pt(**refs),
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.eq({self.sub_equation_parts[0].pt_str(**refs)}, {self.sub_equation_parts[1].pt_str(**refs)})"


class ne(reno.components.Operation):
    """a != b"""

    OP_REPR = "!="

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} \\neq {self.sub_equation_parts[1].latex(**kwargs)}"

    def get_type(self) -> type:
        return bool

    def op_eval(self, **kwargs):
        a, b = adjust_shapes_for_n(*self.sub_equation_parts[0:2], **kwargs)
        return a != b

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.neq(
            self.sub_equation_parts[0].pt(**refs),
            self.sub_equation_parts[1].pt(**refs),
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.ne({self.sub_equation_parts[0].pt_str(**refs)}, {self.sub_equation_parts[1].pt_str(**refs)})"


# TODO: rename to just and?
class bool_and(reno.components.Operation):
    """a and b"""

    OP_REPR = "and"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} \\text{{ and }} {self.sub_equation_parts[1].latex(**kwargs)}"

    def get_type(self) -> type:
        return bool

    def op_eval(self, **kwargs):
        a, b = adjust_shapes_for_n(*self.sub_equation_parts[0:2], **kwargs)
        return a & b

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) & self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} & {self.sub_equation_parts[1].pt_str(**refs)})"


# TODO: rename to just or?
class bool_or(reno.components.Operation):
    """a or b"""

    OP_REPR = "or"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} \\text{{ or }} {self.sub_equation_parts[1].latex(**kwargs)}"

    def get_type(self) -> type:
        return bool

    def op_eval(self, **kwargs):
        a, b = adjust_shapes_for_n(*self.sub_equation_parts[0:2], **kwargs)
        return a | b

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) | self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} | {self.sub_equation_parts[1].pt_str(**refs)})"


class minimum(reno.components.Operation):
    """Element-wise minimum of array elements between two arrays or values, same as np.minimum."""

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"\\text{{min}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        a, b = adjust_shapes_for_n(*self.sub_equation_parts[0:2], **kwargs)
        return np.minimum(a, b)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.minimum(
            self.sub_equation_parts[0].pt(**refs),
            self.sub_equation_parts[1].pt(**refs),
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.minimum({self.sub_equation_parts[0].pt_str(**refs)}, {self.sub_equation_parts[1].pt_str(**refs)})"


class maximum(reno.components.Operation):
    """Element-wise maximum of array elements between two arrays or values, same as np.maximum."""

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"\\text{{max}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        a, b = adjust_shapes_for_n(*self.sub_equation_parts[0:2], **kwargs)
        return np.maximum(a, b)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.maximum(
            self.sub_equation_parts[0].pt(**refs),
            self.sub_equation_parts[1].pt(**refs),
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.maximum({self.sub_equation_parts[0].pt_str(**refs)}, {self.sub_equation_parts[1].pt_str(**refs)})"


class clip(reno.components.Operation):
    """Simultaneously apply upper and lower bound constraint (element-wise)."""

    def __init__(self, a, b, c):
        super().__init__(a, b, c)

    def latex(self, **kwargs):
        return f"\\text{{clip}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)}, {self.sub_equation_parts[2].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        return np.clip(
            self.sub_equation_parts[0].eval(**kwargs),
            self.sub_equation_parts[1].eval(**kwargs),
            self.sub_equation_parts[2].eval(**kwargs),
        )

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.clip(
            self.sub_equation_parts[0].pt(**refs),
            self.sub_equation_parts[1].pt(**refs),
            self.sub_equation_parts[2].pt(**refs),
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.clip({self.sub_equation_parts[0].pt_str(**refs)}, {self.sub_equation_parts[1].pt_str(**refs)}, {self.sub_equation_parts[2].pt_str(**refs)})"


class log(reno.components.Operation):
    """ln(a) (natural log, naming it log because this is pytensor's and numpy's default)"""

    def __init__(self, a):
        super().__init__(a)

    def get_type(self) -> type:
        return float

    def latex(self, **kwargs):
        return f"\\text{{ln}}({self.sub_equation_parts[0].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        return np.log(self.sub_equation_parts[0].eval(**kwargs))

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.log(self.sub_equation_parts[0].pt(**refs))

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.log({self.sub_equation_parts[0].pt_str(**refs)})"


class sin(reno.components.Operation):
    """sin(a)"""

    def __init__(self, a):
        super().__init__(reno.utils.ensure_scalar(a))

    def get_type(self) -> type:
        return float

    def latex(self, **kwargs):
        return f"\\text{{sin}}({self.sub_equation_parts[0].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        return np.sin(self.sub_equation_parts[0].eval(**kwargs))

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.sin(self.sub_equation_parts[0].pt(**refs))

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.sin({self.sub_equation_parts[0].pt_str(**refs)})"


# TODO: all the other trig functions


class interpolate(reno.components.Operation):
    """Given a dataset of x -> y datapoints, interpolate any new data along the line formed by the points.
    Equivalent to numpy's interp function.

    Args:
        x: The input x-coordinates that you want interpolated into y outputs.
        x_data: The x-coordinate data to base interpolation on.
        y_data: the y-coordinate data to base interpolation on.
    """

    def __init__(self, x, x_data: list | np.ndarray, y_data: list | np.ndarray):
        super().__init__(x, x_data, y_data)

    def get_shape(self) -> int:
        """The shapes of x_data and y_data don't matter, should be same as input."""
        return self.sub_equation_parts[0].shape

    def latex(self, **kwargs):
        return f"\\text{{interpolate}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)}, {self.sub_equation_parts[2].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        return np.interp(
            self.sub_equation_parts[0].eval(**kwargs),
            self.sub_equation_parts[1].eval(**kwargs),
            self.sub_equation_parts[2].eval(**kwargs),
        )

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.interpolate1d(
            self.sub_equation_parts[1].pt(**refs),
            self.sub_equation_parts[2].pt(**refs),
            extrapolate=False,
        )(self.sub_equation_parts[0].pt(**refs))

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.interpolate1d({self.sub_equation_parts[1].pt_str(**refs)}, {self.sub_equation_parts[2].pt_str(**refs)},)({self.sub_equation_parts[0].pt_str(**refs)})"


class assign(reno.components.Operation):
    """This is to handle the weird seek_refs issues when you just set a tracked ref's
    equation to another tracked ref. By "wrapping" it in an effectively blank operation,
    this mitigates the annoying recursion issue."""

    OP_REPR = "="

    def __init__(self, a):
        super().__init__(a)

    def get_shape(self) -> int:
        return self.sub_equation_parts[0].shape

    def get_type(self) -> type:
        return self.sub_equation_parts[0].dtype

    def latex(self, **kwargs):
        return self.sub_equation_parts[0].latex(**kwargs)

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[0].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs)

    def pt_str(self, **refs: dict[str, str]) -> str:
        return self.sub_equation_parts[0].pt_str(**refs)


class astype(reno.components.Operation):
    """Allow explicitly converting to float, int, etc."""

    def __init__(self, a, dtype: type):
        self.__dtype = dtype
        super().__init__(a)

    def get_type(self) -> type:
        return self.__dtype

    def latex(self, **kwargs):
        return self.sub_equation_parts[0].latex(**kwargs)

    def op_eval(self, **kwargs):
        sub_val = self.sub_equation_parts[0].eval(**kwargs)
        if isinstance(sub_val, np.ndarray):
            sub_val = sub_val.astype(self.__dtype)
        else:
            sub_val = self.__dtype(sub_val)
        return sub_val

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs).astype(self.__dtype)

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)}).astype({self.__dtype.__name__})"

    @staticmethod
    def parse(arg_strs: list[str], refs: dict[str, reno.components.Reference]):
        if len(arg_strs) != 2:
            raise SyntaxError(
                "astype op expects two arguments, the reference to convert, and the string type to convert to"
            )
        ref = reno.parser.parse(arg_strs[0], refs)
        if arg_strs[1] == "int":
            dtype = int
        elif arg_strs[1] == "float":
            dtype = float
        elif arg_strs[1] == "bool":
            dtype = bool
        return astype(ref, dtype)

    def __repr__(self):
        return (
            f"(astype {self.sub_equation_parts[0].__repr__()} {self.__dtype.__name__})"
        )


class stack(reno.components.Operation):
    """Stack a set of 1 dimensional values into a multidim value.

    This does not yet implement stacking multiple multidim values.
    (may need to be a separate concat operator for that)
    """

    def __init__(self, *args):
        super().__init__(*args)

    def get_shape(self) -> int:
        return len(self.sub_equation_parts)

    def get_type(self) -> type:
        types = [part.dtype for part in self.sub_equation_parts]
        if float in types:
            return float
        return types[0]

    def latex(self, **kwargs) -> str:
        building_str = (
            f"[{','.join(part.latex(**kwargs) for part in self.sub_equation_parts)}]"
        )
        return building_str

    def op_eval(self, **kwargs):
        to_stack = [part.eval(**kwargs) for part in self.sub_equation_parts]

        # have to find the actual n-shape to correctly handle
        implied_n = 1
        for i, part in enumerate(self.sub_equation_parts):
            if (
                part.shape == 1
                and hasattr(to_stack[i], "shape")
                and to_stack[i].shape[0] != 1
            ):
                implied_n = to_stack[i].shape[0]

        for index, item in enumerate(to_stack):
            if implied_n > 1:
                to_stack[index] = np.broadcast_to(item, (implied_n,))

        for i, part in enumerate(to_stack):
            if not isinstance(part, np.ndarray):
                to_stack[i] = np.array([part])
        return np.stack(to_stack, axis=-1)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.stack([part.pt(**refs) for part in self.sub_equation_parts])

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.stack([{','.join(part.pt_str(**refs) for part in self.sub_equation_parts)}])"


# ==================================================
# "HIGHER ORDER"
# ==================================================


class pulse(reno.components.Operation):
    """Return a '1' signal for ``width`` number of timesteps starting at timestep ``start``.
    Returns 0 at all other timesteps."""

    def __init__(self, start, width=1):
        t = reno.components.TimeRef()
        self.sub_eq = reno.components.Piecewise(
            [0, 1],
            [
                (t < start) | (t >= (start + width)),
                (t >= start) & (t < (start + width)),
            ],
        )
        super().__init__(start, width, self.sub_eq)

    def latex(self, **kwargs):
        return f"\\text{{pulse}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[2].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[2].pt(**refs)

    def pt_str(self, **refs: dict[str, str]) -> str:
        return self.sub_equation_parts[2].pt_str(**refs)


class repeated_pulse(reno.components.Operation):
    """Return a '1' signal for ``width`` number of timesteps starting at timestep ``start``,
    with ``interval`` number of timesteps between each subsequent leading edge. Returns 0
    at all other timesteps.

    Example:
        .. code-block:: python

            >>> import reno as r
            >>> m = r.Model()
            >>> m.v0 = r.Variable(r.repeated_pulse(4, 4, 2))
            >>> m()
            >>> m.v0.value
            array([[0, 0, 0, 0, 1, 1, 0, 0, 1, 1]])

            >>> m.v0.eq = r.repeated_pulse(0, 3, 1)
            >>> m()
            >>> m.v0.value
            array([[1, 0, 0, 1, 0, 0, 1, 0, 0, 1]])
    """

    def __init__(self, start, interval, width=1):
        t = reno.components.TimeRef()
        self.sub_eq = reno.components.Piecewise(
            [0, 1],
            [
                (t < start) | ((t - start) % interval >= width),
                (t >= start) & ((t - start) % interval < width),
            ],
        )
        super().__init__(start, interval, width, self.sub_eq)
        # super().__init__(start, interval, width)
        self.non_repr_part_indices.append(3)  # ignore sub_eq in parsing/str

    def latex(self, **kwargs):
        return f"\\text{{repeated\\_pulse}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)}, {self.sub_equation_parts[2].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        # return self.sub_equation_parts[3].eval(**kwargs)
        return self.sub_eq.eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        # return self.sub_equation_parts[3].pt(**refs)
        return self.sub_eq.pt(**refs)

    def pt_str(self, **refs: dict[str, str]) -> str:
        # return self.sub_equation_parts[3].pt_str(**refs)
        return self.sub_eq.pt_str(**refs)


class step(reno.components.Operation):
    """Return a specified value after a specified number of timesteps, otherwise 0."""

    def __init__(self, value, timesteps):
        t = reno.components.TimeRef()
        self.sub_eq = reno.components.Piecewise(
            [0, value], [t < timesteps, t >= timesteps]
        )
        super().__init__(value, timesteps, self.sub_eq)

    def latex(self, **kwargs):
        return f"\\text{{step}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[2].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[2].pt(**refs)

    def pt_str(self, **refs: dict[str, str]) -> str:
        return self.sub_equation_parts[2].pt_str(**refs)


class delay1(reno.components.ExtendedOperation):
    def __init__(self, input, delay_time):
        self.inflow = reno.Flow(input)
        self.delay_stock = reno.Stock()
        self.outflow = reno.Flow(self.delay_stock / delay_time)
        self.delay_stock += self.inflow
        self.delay_stock -= self.outflow
        super().__init__(
            [input, delay_time],
            {
                "inflow": self.inflow,
                "delay_stock": self.delay_stock,
                "outflow": self.outflow,
            },
        )

    def latex(self, **kwargs):
        return f"\\text{{delay1}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        return self.outflow.eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.outflow.pt(**refs)

    def pt_str(self, **refs: dict[str, str]) -> str:
        return self.outflow.pt_str(**refs)


class delay3(reno.components.ExtendedOperation):
    def __init__(self, input, delay_time):
        self.delay1 = delay1(input, delay_time / 3)
        self.delay2 = delay1(self.delay1, delay_time / 3)
        self.delay3 = delay1(self.delay2, delay_time / 3)
        super().__init__([input, delay_time, self.delay3], {})
        # don't need to pass any implicit components up because the individual
        # delays should handle that.

    def latex(self, **kwargs):
        return f"\\text{{delay3}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)})"

    # TODO: does this work just as delay3 or do we need to explicitly reference
    # outflow?
    def op_eval(self, **kwargs):
        return self.delay3.eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.delay3.pt(**refs)

    def pt_str(self, **refs: dict[str, str]) -> str:
        return self.delay3.pt_str(**refs)


class smooth(reno.components.ExtendedOperation):
    """An information delay, material isn't necessarily preserved but the range is?"""

    def __init__(self, input, adjustment_time, initial_value=0):
        self.actual_value = reno.Variable(input, dtype=float)
        self.output = reno.Stock(init=initial_value, dtype=float)  # "perceived" value
        self.gap = reno.Variable(self.actual_value - self.output, dtype=float)
        self.adjustment = reno.Flow(self.gap / adjustment_time, dtype=float)
        self.output += self.adjustment
        super().__init__(
            [input, adjustment_time, initial_value],
            {
                "actual_value": self.actual_value,
                "output": self.output,
                "gap": self.gap,
                "adjustment": self.adjustment,
            },
        )

    def latex(self, **kwargs):
        return f"\\text{{smooth}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)}, {self.sub_equation_parts[2].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        return self.output.eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.output.pt(**refs)

    def pt_str(self, **refs: dict[str, str]) -> str:
        return self.output.pt_str(**refs)


# ==================================================
# META
# ==================================================


class inflow(reno.components.Operation):
    """A purely semantic op to aid in diagramming that allows specifying a flow
    as an "inflow" to another flow. This does not impact any equations, but alters
    rendering to show a thicker "material" line between the two flows.

    This is relevant e.g. when a flow is purely a modification on some other flow,
    and is subsequently used as an inflow to a stock. This op helps avoid the
    separated "islands" that can crop up in certain circumstances"""

    def __init__(self, flow):
        super().__init__(flow)
        if not isinstance(flow, reno.components.Flow):
            raise TypeError(f"The inflow op must be on a Flow, not a {type(flow)}")

    def seek_refs(self, include_ref_types: bool = False):
        if not include_ref_types:
            return [self.sub_equation_parts[0]]
        return {self.sub_equation_parts[0]: ["inflow"]}

    def get_shape(self) -> int:
        return self.sub_equation_parts[0].shape

    def get_type(self) -> type:
        return self.sub_equation_parts[0].dtype

    def latex(self, **kwargs):
        return self.sub_equation_parts[0].latex(**kwargs)

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[0].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs)

    def pt_str(self, **refs: dict[str, str]) -> str:
        return self.sub_equation_parts[0].pt_str(**refs)


class outflows(reno.components.Operation):
    def __init__(self, stock):
        # WAIT: optionally include a list of flows to exclude??
        super().__init__(stock)
        if not isinstance(stock, reno.components.Stock):
            raise TypeError(f"The outflows op must be on a Stock, not a {type(stock)}")

    def seek_refs(self, include_ref_types: bool = False):
        # have to override because outflows is "substituting in" the sum of flows.
        # print("In outflows seek refs!")
        if not include_ref_types:
            return self.sub_equation_parts[0].out_flows
        else:
            return {flow: ["outflows"] for flow in self.sub_equation_parts[0].out_flows}

    def latex(self, **kwargs) -> str:
        return f"{self.sub_equation_parts[0].latex(**kwargs)}.\\text{{outflows}}"

    def op_eval(self, **kwargs):
        return (
            self.sub_equation_parts[0]
            .combine_eqs(self.sub_equation_parts[0].out_flows)
            .eval(**kwargs)
        )

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return (
            self.sub_equation_parts[0]
            .combine_eqs(self.sub_equation_parts[0].out_flows)
            .pt(**refs)
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return (
            self.sub_equation_parts[0]
            .combine_eqs(self.sub_equation_parts[0].out_flows)
            .pt_str(**refs)
        )


# ==================================================
# DISTRIBUTIONS
# ==================================================


def dist_shape(
    dist: reno.components.Distribution, n: int, steps: int, dim: int
) -> int | tuple:
    """Compute the shape/dimensions needed to populate the passed distribution"""
    if not dist.per_timestep and dim == 1:
        shape = n
        # shape = (n, 1)
    elif dist.per_timestep and dim == 1:
        shape = (n, steps)
        # shape = (n, steps, 1)
    elif dist.per_timestep and dim > 1:
        shape = (n, steps, dim)
    else:
        shape = (n, dim)
    return shape


def dist_params(
    dist: reno.components.Distribution, refs: dict
) -> tuple[str, int, str, int]:
    """Extract any dunder variables in the refs passed from pymc.py for setting up parameters
    for the pymc converted distribution."""
    name = "dist" + str(id(dist))
    if "__PTNAME__" in refs:
        name = refs["__PTNAME__"]
    dim = 1
    if "__DIM__" in refs:
        dim = refs["__DIM__"]
    dim_name = "vec"
    if "__DIMNAME__" in refs:
        dim_name = refs["__DIMNAME__"]
    seq = 0
    if "__PT_SEQ_LEN__" in refs:
        seq = refs["__PT_SEQ_LEN__"]
    return name, dim, dim_name, seq


def dist_dim_shape_args(dist: reno.components.Distribution, refs: dict) -> dict:
    name, dim, dim_name, seq = dist_params(dist, refs)
    if not dist.per_timestep and dim == 1:
        return dict()
    elif dist.per_timestep and dim == 1:
        return dict(shape=(seq,), dims="t")
    elif dist.per_timestep and dim > 1:
        return dict(shape=(seq, dim), dims=("t", dim_name))
    else:
        return dict(shape=(dim,), dims=dim_name)


def dist_dim_shape_args_str(
    dist: reno.components.Distribution, refs: dict[str, str]
) -> str:
    name, dim, dim_name, seq = dist_params(dist, refs)
    if not dist.per_timestep and dim == 1:
        return ""
    elif dist.per_timestep and dim == 1:
        return f', shape=({seq},), dims="t"'
    elif dist.per_timestep and dim > 1:
        return f', shape=({seq},, {dim}), dims=("t", "{dim_name}")'
    else:
        return f', shape=({dim},), dims="{dim_name}"'


class Normal(reno.components.Distribution):
    def __init__(self, mean, std=1.0, per_timestep: bool = False):
        # super().__init__()
        # self.mean = mean
        # self.std = std
        super().__init__(mean, std, per_timestep=per_timestep)

    def latex(self, **kwargs):
        return f"\\mathcal{{N}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)}^2)"

    def populate(self, n: int, steps: int = 0, dim: int = 1):
        dims = dist_shape(self, n, steps, dim)
        self.value = np.random.normal(
            self.sub_equation_parts[0].eval(),
            self.sub_equation_parts[1].eval(),
            size=dims,
        )

    def get_type(self) -> type:
        return float

    def __repr__(self):
        if not self.per_timestep:
            return f"Normal({self.clean_part_repr(0)}, {self.clean_part_repr(1)})"
        return f"Normal({self.clean_part_repr(0)}, {self.clean_part_repr(1)}, {self.per_timestep})"

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        name, dim, dim_name, seq = dist_params(self, refs)
        return pm.Normal(
            name,
            self.sub_equation_parts[0].pt(**refs),
            self.sub_equation_parts[1].pt(**refs),
            **dist_dim_shape_args(self, refs),
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        name, dim, dim_name, seq = dist_params(self, refs)
        return f'pm.Normal("{name}", {self.sub_equation_parts[0].pt_str(**refs)}, {self.sub_equation_parts[1].pt_str(**refs)}{dist_dim_shape_args_str(self, refs)})'


class Uniform(reno.components.Distribution):
    def __init__(self, low=0.0, high=1.0, per_timestep: bool = False):
        super().__init__(low, high, per_timestep=per_timestep)

    def latex(self, **kwargs):
        return f"\\mathcal{{U}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)})"

    def get_type(self) -> type:
        return float

    def populate(self, n: int, steps: int = 0, dim: int = 1):
        dims = dist_shape(self, n, steps, dim)
        self.value = np.random.uniform(
            self.sub_equation_parts[0].eval(),
            self.sub_equation_parts[1].eval(),
            size=dims,
        )

    def __repr__(self):
        if not self.per_timestep:
            return f"Uniform({self.clean_part_repr(0)}, {self.clean_part_repr(1)})"
        return f"Uniform({self.clean_part_repr(0)}, {self.clean_part_repr(1)}, {self.per_timestep})"

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        name, dim, dim_name, seq = dist_params(self, refs)
        return pm.Uniform(
            name,
            self.sub_equation_parts[0].pt(**refs),
            self.sub_equation_parts[1].pt(**refs),
            **dist_dim_shape_args(self, refs),
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        name, dim, dim_name, seq = dist_params(self, refs)
        return f'pm.Uniform("{name}", {self.sub_equation_parts[0].pt_str(**refs)}, {self.sub_equation_parts[1].pt_str(**refs)}{dist_dim_shape_args_str(self, refs)})'


class DiscreteUniform(reno.components.Distribution):
    """Low is inclusive, high is exclusive."""

    def __init__(self, low: int = 0, high: int = 2, per_timestep: bool = False):
        super().__init__(low, high, per_timestep=per_timestep)

    def latex(self, **kwargs):
        return f"\\text{{DiscreteUniform}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)})"

    def get_type(self) -> type:
        return int

    def populate(self, n: int, steps: int = 0, dim: int = 1):
        dims = dist_shape(self, n, steps, dim)
        self.value = np.random.randint(
            self.sub_equation_parts[0].eval(),
            self.sub_equation_parts[1].eval(),
            size=dims,
        )

    def __repr__(self):
        if not self.per_timestep:
            return (
                f"DiscreteUniform({self.clean_part_repr(0)}, {self.clean_part_repr(1)})"
            )
        return f"DiscreteUniform({self.clean_part_repr(0)}, {self.clean_part_repr(1)}, {self.per_timestep})"

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        name, dim, dim_name, seq = dist_params(self, refs)
        return pm.DiscreteUniform(
            name,
            self.sub_equation_parts[0].pt(**refs),
            self.sub_equation_parts[1].pt(**refs),
            **dist_dim_shape_args(self, refs),
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        name, dim, dim_name, seq = dist_params(self, refs)
        return f'pm.DiscreteUniform("{name}", {self.sub_equation_parts[0].pt_str(**refs)}, {self.sub_equation_parts[1].pt_str(**refs)}{dist_dim_shape_args_str(self, refs)})'


class Bernoulli(reno.components.Distribution):
    """Discrete single event probability (p is probability of eval == 1)"""

    def __init__(self, p: float, use_p_dist: bool = False, per_timestep: bool = False):
        super().__init__(p, per_timestep=per_timestep)
        self.use_p_dist = use_p_dist

    def latex(self, **kwargs):
        return f"\\text{{Bernoulli}}({self.sub_equation_parts[0].latex(**kwargs)})"

    def populate(self, n: int, steps: int = 0, dim: int = 1):
        dims = dist_shape(self, n, steps, dim)
        self.value = np.random.binomial(1, self.sub_equation_parts[0].eval(), dims)

    def get_type(self) -> type:
        return int

    def __repr__(self):
        if not self.use_p_dist and not self.per_timestep:
            return f"Bernoulli({self.clean_part_repr(0)})"
        return f"Bernoulli({self.clean_part_repr(0)}, {self.use_p_dist}, {self.per_timestep})"

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        name, dim, dim_name, seq = dist_params(self, refs)

        inner_dist = self.sub_equation_parts[0].pt(**refs)
        if self.use_p_dist:
            inner_dist = pm.Interpolated(
                f"{name}_p",
                x_points=np.array([0.0, 1.0]),
                pdf_points=np.array(
                    [
                        1 - self.sub_equation_parts[0].pt(**refs),
                        self.sub_equation_parts[0].pt(**refs),
                    ]
                ),
            )
        return pm.Bernoulli(name, inner_dist, **dist_dim_shape_args(self, refs))

    def pt_str(self, **refs: dict[str, str]) -> str:
        name, dim, dim_name, seq = dist_params(self, refs)

        inner_dist = self.sub_equation_parts[0].pt_str(**refs)
        if self.use_p_dist:
            inner_dist = f'pm.Interpolated("{name}_p", x_points=np.array([0.0, 1.0]), pdf_points=np.array([{1 - self.sub_equation_parts[0].pt_str(**refs)}, {self.sub_equation_parts[0].pt_str(**refs)}]))'
        return (
            f'pm.Bernoulli("{name}", {inner_dist}{dist_dim_shape_args_str(self, refs)})'
        )


class Categorical(reno.components.Distribution):
    """Random categorical distribution - you specify the probability per category,
    and the output is a set of category indices."""

    def __init__(
        self, p: list[float], use_p_dist: bool = False, per_timestep: bool = False
    ):
        super().__init__(p, per_timestep=per_timestep)
        self.use_p_dist = use_p_dist
        self.expected_arg_num_dims[0] = 1

    def latex(self, **kwargs):
        return f"\\text{{Categorical}}({self.sub_equation_parts[0].latex(**kwargs)})"

    def get_type(self) -> type:
        return int

    def populate(self, n: int, steps: int = 0, dim: int = 1):
        dims = dist_shape(self, n, steps, dim)
        # TODO: how would p_dist apply here? Should it?
        self.value = np.argmax(
            np.random.multinomial(1, self.sub_equation_parts[0].eval(), dims), axis=-1
        )

    def __repr__(self):
        return f"Categorical({self.clean_part_repr(0)}, {self.use_p_dist}, {self.per_timestep})"

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        name, dim, dim_name, seq = dist_params(self, refs)

        inner_dist = self.sub_equation_parts[0].pt(**refs)
        if self.use_p_dist:
            inner_dist = pm.Dirichlet(
                f"{name}_p", self.sub_equation_parts[0].pt(**refs)
            )
        return pm.Categorical(name, inner_dist, **dist_dim_shape_args(self, refs))

    def pt_str(self, **refs: dict[str, str]) -> str:
        name, dim, dim_name, seq = dist_params(self, refs)
        inner_dist = self.sub_equation_parts[0].pt_str(**refs)
        if self.use_p_dist:
            inner_dist = f'pm.Dirichlet("{name}_p", {self.p.pt_str(**refs)})'
        return f'pm.Categorical("{name}", {inner_dist}{dist_dim_shape_args_str(self, refs)})'


class List(reno.components.Distribution):
    """Tile passed list to the sample size so each value is hit roughly
    equally (dependent on exact sample size) and deterministically"""

    def __init__(self, values: list | np.ndarray | set):
        super().__init__()
        self.values = values

    def latex(self, **kwargs):
        return f"{self.values}"

    def get_shape(self):
        if isinstance(self.values, np.ndarray) and len(self.values.shape) == 2:
            return self.values.shape[1]
        if isinstance(self.values[0], (set, list)):
            return len(self.values[0])
        return 1

    def get_type(self) -> type:
        # TODO: this doesn't handle a value with a np array with shape > 2
        value = self.values
        # get to the first real value so we can determine type
        while isinstance(value, np.ndarray):
            value = value[0]
        while isinstance(value, (list, set)):
            value = value[0]

        if isinstance(value, np.floating):
            return float
        elif isinstance(value, np.integer):
            return int
        elif isinstance(value, bool):
            return bool
        return type(value)

    def populate(self, n: int, steps: int = 0, dim: int = 1):
        repetitions = n / len(self.values)
        # if the specified value is _larger_ than the samples, we have to
        # truncate (and warn, does the user know this is what's happening?)
        if repetitions < 1:
            warnings.warn(
                f"Not enough samples in simulation to hit every value in list '{self.values}', would require at least `n={len(self.values)}`",
                RuntimeWarning,
            )
            expanded = np.array(self.values)
        else:
            expanded = np.tile(self.values, math.ceil(repetitions))
        self.value = expanded[:n]
        if dim > 1:
            if (
                isinstance(self.values, np.ndarray) and len(self.values.shape) == 1
            ) or not isinstance(self.values[0], (list, set)):
                self.value = np.repeat(np.expand_dims(self.value, axis=1), dim, axis=1)

    def __repr__(self):
        return f"List({self.values})"


class Observation(reno.components.Distribution):
    """Represents a Normal distribution around an observed value.

    Should only be used for supplying observational data with likelihoods
    to bayesian models constructed with model.pymc()

    Args:
        ref (reno.components.Reference): The equation to supply an observed value for.
        sigma (float): The std dev to use for the likelihood Normal distribution.
        data (list): The actual observed data to apply.
    """

    def __init__(
        self,
        ref: reno.components.Reference | reno.components.EquationPart,
        sigma: float = 1.0,
        data: list = None,
    ):
        super().__init__()
        self.ref = ref
        self.sigma = sigma
        self.data = data

    def add_tensors(self, pymc_model):
        with pymc_model:
            # sigma = pm.HalfNormal(f"{self.ref.qual_name()}_sigma", self.sigma)
            pm.Normal(
                f"{self.ref.qual_name()}_likelihood",
                pymc_model[self.ref.qual_name()],
                self.sigma,
                observed=self.data,
            )

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pm.Normal(
            f"{self.ref.qual_name()}_likelihood",
            self.ref.pt(**refs),
            self.sigma,
            observed=self.data,
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f'pm.Normal("{self.ref.qual_name()}_likelihood", {self.ref.pt_str(**refs)}, {self.sigma}, observed={self.data})'


# class Sweep(reno.components.Distribution):
#     """Similar in principle to ops.List, unimplemented idea for this is to
#     grab all variables that are sweeps and collectively make sure they are fully
#     permutated, rather than requiring user to manually make sure they change at
#     varying rates between samples."""
#
#     # TODO: semantically tells to ensure that for _all_ sweep dists
#     # in a system, make sure all value combinations are hit (if size allows, warn if
#     # doesn't)
#     def __init__(self, sweep_values: list | np.ndarray | set):
#         super().__init__()
#         self.sweep_values = sweep_values
#
#     def latex(self, **kwargs):
#         return f"\\text{{Sweep}}({self.sweep_values})"
#
#     def populate(self, n):
#         # TODO: this will be difficult to implement, need to find all other
#         # sweeps in system? We have no ref to that
#         pass
#
#     def __repr__(self):
#         return f"Sweep({self.sweep_values})"
#



"""Outputs from previous runs should probably still continue to be correct."""

import json

import xarray as xr

from reno import ops
from reno.components import Flow, Scalar, Stock, Variable
from reno.model import Model


def test_bottleneck_model():
    """This model was used to help understand how min/max's interact with throughputs,
    e.g. how do you correctly limit flow mins/maxes to ensure the bottleneck is retained,
    but you don't end up with a really jagged/timestep-oscilating output?

    The xarray datasets used for this are generated from notebook 14_bottleneck_example
    """

    m = Model()
    m.s1 = Stock(doc="Input material stockpile")
    m.s2 = Stock(doc="Internal processing facility temp stock")
    m.s3 = Stock(doc="Output material stockpile")
    m.f1 = Flow()
    m.f2 = Flow()
    m.f3 = Flow(doc="bottleneck that needs a lower throughput than this one")

    m.s2_max = Variable()
    m.process_throughput = Variable()
    m.s3_max = Variable()

    m.s1 -= m.f1
    m.s2 += m.f1
    m.s2 -= m.f2
    m.s3 += m.f2
    m.s3 -= m.f3

    m.f1.eq = m.process_throughput
    m.f1.max = ops.minimum(m.s1, m.s2_max - m.s2)
    m.f1.min = Scalar(0)

    m.f2.eq = m.process_throughput
    m.f2.max = ops.minimum(m.s2, m.s3_max - m.s3)
    m.f2.min = Scalar(0)

    m.f3.eq = m.process_throughput / 2
    m.f3.max = m.s3
    m.f3.min = Scalar(0)

    ds1 = m(s1_0=100, s2_max=10, process_throughput=10, s3_max=30)
    ds2 = m(s1_0=100, s2_max=20, process_throughput=10, s3_max=30)
    ds3 = m(s1_0=100, s2_max=30, process_throughput=10, s3_max=30)
    ds4 = m(s1_0=100, s2_max=15, process_throughput=10, s3_max=30)

    # ds1 = json.dumps(ds1.to_dict(), indent=4, default=lambda x: str(x))
    # ds2 = json.dumps(ds2.to_dict(), indent=4, default=lambda x: str(x))
    # ds3 = json.dumps(ds3.to_dict(), indent=4, default=lambda x: str(x))
    # ds4 = json.dumps(ds4.to_dict(), indent=4, default=lambda x: str(x))
    # ds1 = ds1.to_dict()
    # ds2 = ds2.to_dict()
    # ds3 = ds3.to_dict()
    # ds4 = ds4.to_dict()

    with open("tests/regression_data/ds1.json") as infile:
        ds1_prev = xr.Dataset.from_dict(json.load(infile))
    with open("tests/regression_data/ds2.json") as infile:
        ds2_prev = xr.Dataset.from_dict(json.load(infile))
    with open("tests/regression_data/ds3.json") as infile:
        ds3_prev = xr.Dataset.from_dict(json.load(infile))
    with open("tests/regression_data/ds4.json") as infile:
        ds4_prev = xr.Dataset.from_dict(json.load(infile))

    assert ds1.equals(ds1_prev)
    # assert ds1 == ds1_prev
    assert ds2 == ds2_prev
    assert ds3 == ds3_prev
    assert ds4 == ds4_prev



"""Make sure index selection works on the interactive latex selector."""

import pytest

from reno.viz import ModelLatex


@pytest.mark.parametrize(
    "show_docs,start_name,stop_name,i,expected_name",
    [
        (False, None, None, 1, "faucet_volume"),
        (True, None, None, 1, "faucet_shutoff_time"),
        (False, "faucet_volume", "drain", 0, "faucet"),
        (False, "faucet_volume", "drain", 1, "drain"),
        (False, "faucet_volume", "drain", 2, None),
    ],
)
def test_equation_index(tub_model, show_docs, start_name, stop_name, i, expected_name):
    """Requesting an equation name via index with varying start/stop conditions and doc statuses should return the correct name."""
    latex = ModelLatex(tub_model, show_docs, start_name, stop_name)
    assert latex.find_equation_name_from_index(i) == expected_name


def test_lines_refname_reference(tub_model):
    """The equation lines aligned refname list should be correct."""
    latex = ModelLatex(tub_model, show_docs=True)
    assert latex._equation_lines_refname_reference() == [
        "faucet_shutoff_time",
        "faucet_shutoff_time",
        "faucet_volume",
        "faucet_volume",
        "faucet",
        "drain",
        "water_level",
        "water_level",
        "water_level",
        "final_water_level",
    ]



"""
System dynamics library
=======================

Library for constructing, debugging, visualizing, and analyzing system
dynamics models and simulations.

>>> from reno.components import Stock, Flow, Variable, Scalar, TimeRef
>>> from reno.model import Model
>>> from reno import ops
>>> t = TimeRef()
>>> tub = Model("tub")
>>> tub.faucet = Flow(Scalar(5))
>>> tub.drain = Flow(ops.sin(t) + 2)
>>> tub.water_level = Stock()
>>> tub.water_level += tub.faucet
>>> tub.water_level -= tub.drain
>>> tub()
"""

import warnings

import reno.components  # noqa: F401
import reno.diagrams  # noqa: F402
import reno.explorer  # noqa: F401
import reno.model  # noqa: F401
import reno.ops  # noqa: F401
import reno.pymc  # noqa: F401
import reno.viz  # noqa: F401
from reno.components import Flag  # noqa: F401
from reno.components import Flow  # noqa: F401
from reno.components import Function  # noqa: F401
from reno.components import Metric  # noqa: F401
from reno.components import Piecewise  # noqa: F401
from reno.components import Scalar  # noqa: F401
from reno.components import Stock  # noqa: F401
from reno.components import TimeRef  # noqa: F401
from reno.components import Variable  # noqa: F401
from reno.model import Model  # noqa: F401
from reno.ops import *  # noqa: F401, F403
from reno.viz import plot_refs  # noqa: F401
from reno.viz import plot_refs_single_axis  # noqa: F401
from reno.viz import plot_trace_refs  # noqa: F401

warnings.simplefilter("always", RuntimeWarning)

__version__ = "0.10.0"



"""Script to export clean copies of interactive explorer visualizations/session tabs."""

import base64
import json
import os
import sys

import panel as pn
from bokeh.resources import INLINE

import reno


def load_pn_panes(model_file: str, panesset_file: str) -> reno.explorer.PanesSet:
    """Re-populate a new tab of panes from a JSON serialized panes file.

    Args:
        model_file (str): Path to a model JSON file.
        panesset_file (str): Path to a saved tab of panes (see explorer.py)

    Returns:
        A PanesSet object with all the same panes and data originally saved
        into panesset_file.
    """
    pn.extension("gridstack", "texteditor", "terminal")

    with open(model_file) as infile:
        model_data = json.load(infile)

    with open(panesset_file) as infile:
        panesset_data = json.load(infile)

    model = reno.model.Model.from_dict(model_data)

    panes = reno.explorer.PanesSet(model)
    panes.from_dict(panesset_data)

    return panes


def export_clean_html(model_file: str, panesset_file: str, output_path: str):
    """Cleaner approach to export_bokeh_html. This relies on every pane type having a
    ``to_html`` function that outputs something nice and static, (e.g. embedding images,
    little to no JS, etc.).

    Resulting HTML should be standalone and is much more amenable to printing/PDF export.

    Args:
        model_file (str): Path to a model JSON file.
        panesset_file (str): Path to a saved tab of panes (see explorer.py)
        output_path (str): Path to write the output HTML (include filename.)
    """
    paneset = load_pn_panes(model_file, panesset_file)

    for pane in paneset.panes:
        if hasattr(pane, "render"):
            pane.render(paneset.active_traces)

    html = """
    <head>
        <style>
            @page {
                size: A4 portrait;
                margin: .05in;
            }
            .grid-container {
                display: grid;
            }
            img {
                object-fit: scale-down;
                max-width: 100%;
                height: auto;
            }
        </style>
    </head>
    <body>
        <div class="grid-container">
    """

    for loc, obj in paneset.gstack.objects.items():
        html += f"<div class='grid-cell' style='grid-row-start: {loc[0] + 1}; grid-column-start: {loc[1] + 1}; grid-row-end: {loc[2] + 1}; grid-column-end: {loc[3] + 1};'>"
        html += obj.to_html()
        html += "</div>"

    html += "</div></body>"

    with open(output_path, "w") as outfile:
        outfile.write(html)


def export_bokeh_html(model_file: str, panesset_file: str, output_path: str):
    """Leaving in just in case we find it useful, but this effectively dumps out a small HTML
    that just loads in a ****ton of JS to populate every bokeh component. This is fine for
    just HTML outputs, but very messy to try to create printable PDFs with."""
    paneset = load_pn_panes(model_file, panesset_file)

    # simplify objects a little
    grid = pn.GridSpec(height=paneset.gstack.height, sizing_mode="stretch_both")
    grid.objects = {
        key: value.clicker.child for key, value in paneset.gstack.objects.items()
    }
    objs = [value.clicker.child for key, value in paneset.gstack.objects.items()]

    # so frustrating, a blank panel repo saving stuff doesn't have
    # any visbility issues, but for some reason in this project it
    # exports invisible every single time...so here's a hack to deal
    # with it for now
    force_visible = """
    :root, :host, :host div div, :root div div, div {
        visibility: visible !important;
    }
    """
    for obj in objs:
        obj.visible = True
        obj.stylesheets.append(force_visible)
        if hasattr(obj, "objects"):
            for subobj in obj.objects:
                subobj.visible = True
                subobj.stylesheets.append(force_visible)

    for pane in paneset.panes:
        if hasattr(pane, "render"):
            pane.render(paneset.active_traces)

    grid.save(output_path, resources=INLINE, embed=True)


def export_selenium_pdf(input_path: str, output_path: str):
    """Create a PDF from an HTML file using selenium's chrome webdriver.

    Args:
        input_path (str): The filepath/name of the HTML to render.
        output_path (str): The filepath/name to save the output PDF at.
    """
    from selenium import webdriver
    from selenium.webdriver.chrome.options import Options
    from selenium.webdriver.common.print_page_options import PrintOptions

    options = Options()
    options.add_argument("--headless")
    driver = webdriver.Chrome(options=options)
    driver.get(f"file://{os.path.abspath(input_path)}")
    print_options = PrintOptions()
    print_options.shink_to_fit = True
    pdf = base64.b64decode(driver.print_page(print_options))
    with open(output_path, "wb") as outfile:
        outfile.write(pdf)


if __name__ == "__main__":
    model_file_path = sys.argv[1]
    panes_file_path = sys.argv[2]
    html_file_path = sys.argv[3]
    pdf_file_path = sys.argv[4]

    export_clean_html(model_file_path, panes_file_path, html_file_path)
    export_selenium_pdf(html_file_path, pdf_file_path)
